"""Memory-backed store implementations
"""
from collections import OrderedDict
from copy import copy, deepcopy

from smif.data_layer.abstract_config_store import ConfigStore
from smif.data_layer.abstract_data_store import DataStore
from smif.data_layer.abstract_metadata_store import MetadataStore
from smif.data_layer.data_array import DataArray
from smif.exception import (SmifDataExistsError, SmifDataMismatchError,
                            SmifDataNotFoundError)


class MemoryConfigStore(ConfigStore):
    """Config store in memory
    """
    def __init__(self):
        super().__init__()
        self._model_runs = OrderedDict()
        self._sos_models = OrderedDict()
        self._models = OrderedDict()
        self._scenarios = OrderedDict()
        self._narratives = OrderedDict()
        self._strategies = OrderedDict()

    # region Model runs
    def read_model_runs(self):
        return list(self._model_runs.values())

    def read_model_run(self, model_run_name):
        try:
            return self._model_runs[model_run_name]
        except KeyError:
            raise SmifDataNotFoundError("sos_model_run '%s' not found" % (model_run_name))

    def write_model_run(self, model_run):
        if model_run['name'] not in self._model_runs:
            self._model_runs[model_run['name']] = model_run
        else:
            raise SmifDataExistsError("model_run '%s' already exists" % (model_run['name']))

    def update_model_run(self, model_run_name, model_run):
        if model_run_name in self._model_runs:
            self._model_runs[model_run_name] = model_run
        else:
            raise SmifDataNotFoundError("model_run '%s' not found" % (model_run_name))

    def delete_model_run(self, model_run_name):
        try:
            del self._model_runs[model_run_name]
        except KeyError:
            raise SmifDataNotFoundError("model_run '%s' not found" % (model_run_name))
    # endregion

    # region System-of-systems models
    def read_sos_models(self):
        return list(self._sos_models.values())

    def read_sos_model(self, sos_model_name):
        try:
            return self._sos_models[sos_model_name]
        except KeyError:
            raise SmifDataNotFoundError("sos_model '%s' not found" % (sos_model_name))

    def write_sos_model(self, sos_model):
        if sos_model['name'] not in self._sos_models:
            self._sos_models[sos_model['name']] = sos_model
        else:
            raise SmifDataExistsError("sos_model '%s' already exists" % (sos_model['name']))

    def update_sos_model(self, sos_model_name, sos_model):
        if sos_model_name in self._sos_models:
            self._sos_models[sos_model_name] = sos_model
        else:
            raise SmifDataNotFoundError("sos_model '%s' not found" % (sos_model_name))

    def delete_sos_model(self, sos_model_name):
        try:
            del self._sos_models[sos_model_name]
        except KeyError:
            raise SmifDataNotFoundError("sos_model '%s' not found" % (sos_model_name))
    # endregion

    # region Models
    def read_models(self):
        return list(self._models.values())

    def read_model(self, model_name):
        try:
            return self._models[model_name]
        except KeyError:
            raise SmifDataNotFoundError("model '%s' not found" % (model_name))

    def write_model(self, model):
        if model['name'] not in self._models:
            model = _skip_coords(model, ('inputs', 'outputs', 'parameters'))
            self._models[model['name']] = model
        else:
            raise SmifDataExistsError("model '%s' already exists" % (model['name']))

    def update_model(self, model_name, model):
        if model_name in self._models:
            model = _skip_coords(model, ('inputs', 'outputs', 'parameters'))
            self._models[model_name] = model
        else:
            raise SmifDataNotFoundError("model '%s' not found" % (model_name))

    def delete_model(self, model_name):
        try:
            del self._models[model_name]
        except KeyError:
            raise SmifDataNotFoundError("model '%s' not found" % (model_name))
    # endregion

    # region Scenarios
    def read_scenarios(self):
        scenarios = self._scenarios.values()
        return [_variant_dict_to_list(s) for s in scenarios]

    def read_scenario(self, scenario_name):
        try:
            scenario = self._scenarios[scenario_name]
            return _variant_dict_to_list(scenario)
        except KeyError:
            raise SmifDataNotFoundError("scenario '%s' not found" % (scenario_name))

    def write_scenario(self, scenario):
        if scenario['name'] not in self._scenarios:
            scenario = _variant_list_to_dict(scenario)
            scenario = _skip_coords(scenario, ['provides'])
            self._scenarios[scenario['name']] = scenario
        else:
            raise SmifDataExistsError("scenario '%s' already exists" % (scenario['name']))

    def update_scenario(self, scenario_name, scenario):
        if scenario_name in self._scenarios:
            scenario = _variant_list_to_dict(scenario)
            scenario = _skip_coords(scenario, ['provides'])
            self._scenarios[scenario_name] = scenario
        else:
            raise SmifDataNotFoundError("scenario '%s' not found" % (scenario_name))

    def delete_scenario(self, scenario_name):
        try:
            del self._scenarios[scenario_name]
        except KeyError:
            raise SmifDataNotFoundError("scenario '%s' not found" % (scenario_name))
    # endregion

    # region Scenario Variants
    def read_scenario_variants(self, scenario_name):
        return list(self._scenarios[scenario_name]['variants'].values())

    def read_scenario_variant(self, scenario_name, variant_name):
        try:
            return self._scenarios[scenario_name]['variants'][variant_name]
        except KeyError:
            raise SmifDataNotFoundError("scenario '%s' variant '%s' not found"
                                        % (scenario_name, variant_name))

    def write_scenario_variant(self, scenario_name, variant):
        self._scenarios[scenario_name]['variants'][variant['name']] = variant

    def update_scenario_variant(self, scenario_name, variant_name, variant):
        self._scenarios[scenario_name]['variants'][variant_name] = variant

    def delete_scenario_variant(self, scenario_name, variant_name):
        del self._scenarios[scenario_name]['variants'][variant_name]
    # endregion

    # region Narratives
    def _read_narratives(self, sos_model_name):
        return self._sos_models[sos_model_name]['narratives']

    def read_narrative(self, sos_model_name, narrative_name):
        try:
            narrative = [x for x in self._read_narratives(sos_model_name)
                         if x['name'] == narrative_name][0]
        except IndexError:
            msg = "Narrative '{}' not found in '{}'"
            raise SmifDataNotFoundError(msg.format(narrative_name, sos_model_name))
        return narrative

    def _read_narrative_variant(self, sos_model_name, narrative_name, variant_name):
        narrative = self.read_narrative(sos_model_name, narrative_name)
        try:
            variant = [x for x in narrative['variants'] if x['name'] == variant_name][0]
        except IndexError:
            msg = "Variant '{}' not found in '{}'"
            raise SmifDataNotFoundError(msg.format(variant_name, narrative_name))
        return variant
    # endregion

    # region Strategies
    def read_strategies(self, modelrun_name):
        try:
            return self._strategies[modelrun_name]
        except KeyError:
            raise SmifDataNotFoundError("strategies in modelrun '%s' not found"
                                        % (modelrun_name))

    def write_strategies(self, modelrun_name, strategies):
        self._strategies[modelrun_name] = strategies
    # endregion


class MemoryMetadataStore(MetadataStore):
    """Store metadata in-memory
    """
    def __init__(self):
        super().__init__()
        self._units = []  # list[str] of pint definitions
        self._dimensions = OrderedDict()

    # region Units
    def write_unit_definitions(self, units):
        self._units = units

    def read_unit_definitions(self):
        return self._units
    # endregion

    # region Dimensions
    def read_dimensions(self, skip_coords=False):
        return [self.read_dimension(k, skip_coords) for k in self._dimensions]

    def read_dimension(self, dimension_name, skip_coords=False):
        dim = self._dimensions[dimension_name]
        if skip_coords:
            dim = {
                'name': dim['name'],
                'description': dim['description']
            }
        return dim

    def write_dimension(self, dimension):
        self._dimensions[dimension['name']] = dimension

    def update_dimension(self, dimension_name, dimension):
        self._dimensions[dimension['name']] = dimension

    def delete_dimension(self, dimension_name):
        del self._dimensions[dimension_name]
    # endregion


class MemoryDataStore(DataStore):
    """Store data in-memory
    """
    def __init__(self):
        super().__init__()
        self._data_array = OrderedDict()
        self._interventions = OrderedDict()
        self._initial_conditions = OrderedDict()
        self._state = OrderedDict()
        self._model_parameter_defaults = OrderedDict()
        self._coefficients = OrderedDict()
        self._results = OrderedDict()

    # region Data Array
    def read_scenario_variant_data(self, key, spec, timestep=None):
        return self._read_data_array(key, spec, timestep)

    def write_scenario_variant_data(self, key, data, timestep=None):
        self._write_data_array(key, data, timestep)

    def read_narrative_variant_data(self, key, spec, timestep=None):
        return self._read_data_array(key, spec, timestep)

    def write_narrative_variant_data(self, key, data, timestep=None):
        self._write_data_array(key, data, timestep)

    def _read_data_array(self, key, spec, timestep=None):
        if timestep:
            try:
                data = self._data_array[key, timestep]
            except KeyError:
                try:
                    data = self._filter_timestep(self._data_array[key], spec, timestep)
                except KeyError:
                    raise SmifDataNotFoundError(
                        "Data for {} not found for timestep {}".format(spec.name, timestep))
        else:
            try:
                data = self._data_array[key]
            except KeyError:
                raise SmifDataNotFoundError(
                    "Data for {} not found".format(spec.name))

        if data.spec != spec:
            raise SmifDataMismatchError(
                "Spec did not match reading {}, requested {}, got {}".format(
                    spec.name, spec, data.spec))
        return data

    def _filter_timestep(self, data, read_spec, timestep):
        dataframe = data.as_df().reset_index()
        if 'timestep' not in dataframe.columns:
            msg = "Missing 'timestep' key, found {} in {}"
            raise SmifDataMismatchError(msg.format(list(dataframe.columns), data.name))
        dataframe = dataframe[dataframe.timestep == timestep]
        if dataframe.empty:
            raise SmifDataNotFoundError(
                "Data for {} not found for timestep {}".format(data.name, timestep))
        dataframe.drop('timestep', axis=1, inplace=True)
        return DataArray.from_df(read_spec, dataframe)

    def _write_data_array(self, key, data, timestep=None):
        if timestep:
            self._data_array[key, timestep] = data
        else:
            self._data_array[key] = data
    # endregion

    # region Model parameters
    def read_model_parameter_default(self, key, spec):
        data = self._model_parameter_defaults[key]
        if data.spec != spec:
            raise SmifDataMismatchError(
                "Spec did not match reading {}, requested {}, got {}".format(
                    spec.name, spec, data.spec))
        return data

    def write_model_parameter_default(self, key, data):
        self._model_parameter_defaults[key] = data
    # endregion

    # region Interventions
    def read_interventions(self, keys):
        all_interventions = {}
        interventions = [list(self._interventions[key].values()) for key in keys][0]

        for entry in interventions:
            name = entry.pop('name')
            if name in all_interventions:
                msg = "An entry for intervention {} already exists"
                raise ValueError(msg.format(name))
            else:
                all_interventions[name] = entry

        return all_interventions

    def write_interventions(self, key, interventions):
        self._interventions[key] = interventions

    def read_strategy_interventions(self, strategy):
        return strategy['interventions']

    def read_initial_conditions(self, keys):
        return [self._initial_conditions[key] for key in keys][0]

    def write_initial_conditions(self, key, initial_conditions):
        self._initial_conditions[key] = initial_conditions
    # endregion

    # region State
    def read_state(self, modelrun_name, timestep=None, decision_iteration=None):
        return self._state[(modelrun_name, timestep, decision_iteration)]

    def write_state(self, state, modelrun_name, timestep=None, decision_iteration=None):
        self._state[(modelrun_name, timestep, decision_iteration)] = state
    # endregion

    # region Conversion coefficients
    def read_coefficients(self, source_dim, destination_dim):
        try:
            return self._coefficients[(source_dim, destination_dim)]
        except KeyError:
            msg = "Could not find coefficients for conversion from {}>{}"
            raise SmifDataNotFoundError(msg.format(source_dim, destination_dim))

    def write_coefficients(self, source_dim, destination_dim, data):
        self._coefficients[(source_dim, destination_dim)] = data
    # endregion

    # region Results
    def read_results(self, modelrun_name, model_name, output_spec, timestep=None,
                     decision_iteration=None):
        key = (modelrun_name, model_name, output_spec.name, timestep, decision_iteration)

        try:
            results = self._results[key]
        except KeyError:
            raise SmifDataNotFoundError("Cannot find results for {}".format(key))

        return DataArray(output_spec, results)

    def write_results(self, data_array, modelrun_name, model_name, timestep=None,
                      decision_iteration=None):
        key = (modelrun_name, model_name, data_array.spec.name, timestep, decision_iteration)
        self._results[key] = data_array.as_ndarray()

    def available_results(self, model_run_name):
        results_keys = [
            (timestep, decision_iteration, model_name, output_name)
            for (result_modelrun_name, model_name, output_name, timestep, decision_iteration)
            in self._results.keys()
            if model_run_name == result_modelrun_name
        ]
        return results_keys
    # endregion


def _variant_list_to_dict(config):
    config = copy(config)
    try:
        list_ = config['variants']
    except KeyError:
        list_ = []
    config['variants'] = {variant['name']: variant for variant in list_}
    return config


def _variant_dict_to_list(config):
    config = copy(config)
    try:
        dict_ = config['variants']
    except KeyError:
        dict_ = {}
    config['variants'] = list(dict_.values())
    return config


def _skip_coords(config, keys):
    """Given a config dict and list of top-level keys for lists of specs,
    delete coords from each spec in each list.
    """
    config = deepcopy(config)
    for key in keys:
        for spec in config[key]:
            try:
                del spec['coords']
            except KeyError:
                pass
    return config



"""Data is typically multi-dimensional. :class:`~smif.metadata.spec.Spec` is used to describe
each dataset which is supplied to - or output from - each :class:`~smif.model.model.Model` in a
:class:`~smif.model.model.CompositeModel`
"""
from collections import defaultdict

from smif.metadata.coordinates import Coordinates


class Spec(object):
    """N-dimensional metadata.

    Spec labels each dimension with coordinates and enforces data type, units and absolute and
    expected ranges.

    The API here is modelled on :class:`xarray.DataArray`: dtype and shape describe a
    :class:`numpy.ndarray`; dims and coords follow the xarray conventions for labelled axes;
    and unit, abs_range and exp_range are introduced as supplementary metadata to
    help validate connections between models.

    Attributes
    ----------
    name : str
        The name of the data that this spec describes
    description : str
        A human-friendly description
    dtype : str
        Data type for data values
    abs_range : tuple
        Absolute range of data values
    exp_range : tuple
        Expected range of data values
    shape : tuple[int]
        Tuple of dimension sizes
    ndim : int
        Number of dimensions
    dims : list[str]
        Dimension names
    coords : list[Coordinates]
        Dimension coordinate labels
    unit : str
        Units of data values

    Parameters
    ----------
    name : str, optional
        Name to identifiy the variable described (typically an input, output or parameter)
    description : str, optional
        Short description
    dims : list[str], optional
        List of dimension names, must be provided if coords is a dict
    coords : list[Coordinates] or dict[str, list], optional
        A list of :class`Coordinates` or a dict mapping each dimension name to a list of names
        which label that dimension.
    dtype : str
        String suitable for contructing a simple :class:`numpy.dtype`
    abs_range : tuple, optional
        (min, max) absolute range for numeric values - can be used to raise errors
    exp_range : tuple, optional
        (min, max) expected range for numeric values - can be used to raise warnings
    unit : str, optional
        Unit to be used for data values
    """
    def __init__(self, name=None, dims=None, coords=None, dtype=None,
                 abs_range=None, exp_range=None, unit=None, description=None):
        self._name = name
        self._description = description

        # Coords may come as a dict, in which case dims must be provided to define order
        if isinstance(coords, dict):
            try:
                coords, dims = self._coords_from_dict(coords, dims)
            except (ValueError, KeyError) as error:
                msg = "Coordinate metadata incorrectly formatted for variable '{}': {}"
                raise ValueError(msg.format(self.name, error))

        # Or as a list of Coordinates, in which case dims must not be provided
        elif isinstance(coords, list):
            coords, dims = self._coords_from_list(coords, dims)

        # Or if None, this spec describes a zero-dimensional parameter - single value
        else:
            coords, dims = [], []

        self._dims = dims
        self._coords = coords

        if dtype is None:
            raise ValueError("Spec.dtype must be provided, in {}".format(self._name))
        self._dtype = dtype

        if abs_range is not None:
            self._check_range(abs_range)
        self._abs_range = abs_range

        if exp_range is not None:
            self._check_range(exp_range)
        self._exp_range = exp_range

        self._unit = unit

    def _coords_from_list(self, coords, dims):
        """Set up coords and dims, checking for consistency
        """
        for coord in coords:
            if not isinstance(coord, Coordinates):
                msg = "Spec.coords may be a dict[str,list] or a list[Coordinates], in {}"
                raise ValueError(msg.format(self._name))

        if dims is not None:
            msg = "Spec.dims are derived from Spec.coords if provided as a list of " + \
                  "Coordinates, in {}"
            raise ValueError(msg.format(self._name))

        dims = [coord.dim for coord in coords]

        if len(dims) != len(set(dims)):
            msg = "Spec cannot be created with duplicate dims, in {}"
            raise ValueError(msg.format(self._name))

        return coords, dims

    def _coords_from_dict(self, coords, dims):
        """Set up coords and dims, checking for consistency
        """
        if dims is None:
            msg = "Spec.dims must be specified if coords are provided as a dict, in {}"
            raise ValueError(msg.format(self._name))

        if len(dims) != len(set(dims)):
            msg = "Spec cannot be created with duplicate dims, in {}"
            raise ValueError(msg.format(self._name))

        if sorted(dims) != sorted(coords.keys()):
            msg = "Spec.dims must match the keys in coords, in {}"
            raise ValueError(msg.format(self._name))

        coords = [Coordinates(dim, coords[dim]) for dim in dims]

        return coords, dims

    @classmethod
    def from_dict(cls, data_provided):
        """Create a Spec from a dict representation
        """
        # default anything to None, let constructor handle essential missing values
        data = defaultdict(lambda: None)
        data.update(data_provided)
        spec = Spec(
            name=data['name'],
            description=data['description'],
            dims=data['dims'],
            coords=data['coords'],
            dtype=data['dtype'],
            abs_range=data['abs_range'],
            exp_range=data['exp_range'],
            unit=data['unit']
            )
        return spec

    def as_dict(self):
        """Serialise to dict representation
        """
        return {
            'name': self.name,
            'description': self.description,
            'dims': self._dims,
            'coords': {c.name: c.ids for c in self._coords},
            'dtype': self._dtype,
            'abs_range': self._abs_range,
            'exp_range': self._exp_range,
            'unit': self._unit
        }

    @property
    def name(self):
        """The name of the data that this spec describes.
        """
        return self._name

    @name.setter
    def name(self, value):
        self._name = value

    @property
    def description(self):
        """A human-friendly description
        """
        return self._description

    @property
    def dtype(self):
        """The dtype of the data that this spec describes.
        """
        return self._dtype

    @property
    def abs_range(self):
        """The absolute range of data values that this spec describes.
        """
        return self._abs_range

    @property
    def exp_range(self):
        """The expected range of data values that this spec describes.
        """
        return self._exp_range

    @property
    def shape(self):
        """Tuple of dimension sizes. The shape of the data that this spec describes.
        """
        return tuple(len(c.ids) for c in self._coords)

    @property
    def ndim(self):
        """The number of dimensions of the data that this spec describes.
        """
        return len(self._coords)

    @property
    def dims(self):
        """Names for each dimension
        """
        return list(self._dims)

    @property
    def coords(self):
        """Coordinate labels for each dimension.
        """
        return list(self._coords)

    def dim_coords(self, dim: str):
        """Coordinates for a given dimension
        """
        if not isinstance(dim, str):
            msg = "Expected string as argument, instead received {}"
            raise TypeError(msg.format(type(dim)))

        if dim not in self.dims:
            raise KeyError("Could not find dim '{}' in Spec '{}'".format(dim, self._name))

        for coord in self._coords:
            if coord.dim == dim:
                return coord
        raise KeyError("Coords not found for dim '{}', in Spec '{}'".format(dim, self._name))

    def dim_names(self, dim: str):
        """Names of each coordinate in a given dimension
        """
        return self.dim_coords(dim).names

    def dim_elements(self, dim: str):
        """Elements of each coordinate in a given dimension
        """
        return self.dim_coords(dim).elements

    @property
    def unit(self):
        """The unit for all data points.
        """
        return self._unit

    def __eq__(self, other):
        return self.dtype == other.dtype \
            and self.dims == other.dims \
            and self.coords == other.coords \
            and self.unit == other.unit

    def __hash__(self):
        return hash((
            self.dtype,
            tuple(self.dims),
            tuple(self.coords),
            self.unit
        ))

    def __repr__(self):
        return "<Spec name='{}' dims='{}' unit='{}'>".format(self.name, self.dims, self.unit)

    def _check_range(self, range_):
        """Error if range is not a [min, max] list or tuple
        """
        if not _is_sequence(range_):
            msg = "Spec range must be a list or tuple, got {} for {}"
            raise TypeError(msg.format(range_, self._name))
        if len(range_) != 2:
            msg = "Spec range must have min and max values only, got {} for {}"
            raise ValueError(msg.format(range_, self._name))
        min_, max_ = range_
        if max_ < min_:
            msg = "Spec range min value must be smaller than max value, got {} for {}"
            raise ValueError(msg.format(range_, self._name))


def _is_sequence(obj):
    """Check for iterable object that is not a string ('strip' is a method on str)
    """
    return not hasattr(obj, "strip") \
        and (hasattr(obj, "__getitem__") or hasattr(obj, "__iter__"))



"""Test ModelRunScheduler and JobScheduler
"""
from unittest.mock import Mock, patch

import networkx
from pytest import fixture, raises
from smif.controller.scheduler import JobScheduler, ModelRunScheduler
from smif.model import ModelOperation, ScenarioModel, SectorModel


class EmptySectorModel(SectorModel):
    def simulate(self, data):
        return data


class TestModelRunScheduler():
    @patch('smif.controller.scheduler.subprocess.Popen')
    def test_single_modelrun(self, mock_popen):
        my_scheduler = ModelRunScheduler()
        my_scheduler.add('my_model_run', {
            'directory': 'mock/dir',
            'verbosity': 0,
            'warm_start': False,
            'output_format': 'local_csv'
        })

        mock_popen.assert_called_with(
            'smif  run my_model_run -d mock/dir -i local_csv',
            shell=True,
            stderr=-2, stdout=-1
        )

    def test_status_modelrun_never_added(self):
        my_scheduler = ModelRunScheduler()
        status = my_scheduler.get_status('my_model_run')
        assert status['status'] == 'unstarted'

    @patch('smif.controller.scheduler.subprocess.Popen')
    def test_status_model_started(self, mock_popen):
        attrs = {
            'poll.return_value': None,
            'communicate.return_value': (
                "this is a stdout".encode('utf-8'),
            ),
            'returncode': None
        }
        process_mock = Mock(**attrs)
        mock_popen.return_value = process_mock

        my_scheduler = ModelRunScheduler()
        my_scheduler.add('my_model_run', {
            'directory': 'mock/dir',
            'verbosity': 0,
            'warm_start': False,
            'output_format': 'local_csv'
        })
        my_scheduler.lock = True
        status = my_scheduler.get_status('my_model_run')
        assert status['status'] == 'running'

    @patch('smif.controller.scheduler.subprocess.Popen')
    def test_status_model_done(self, mock_popen):
        attrs = {
            'poll.return_value': 0,
            'communicate.return_value': (
                "this is a stdout".encode('utf-8')
            )
        }
        process_mock = Mock(**attrs)
        mock_popen.return_value = process_mock

        my_scheduler = ModelRunScheduler()
        my_scheduler.add('my_model_run', {
            'directory': 'mock/dir',
            'verbosity': 0,
            'warm_start': False,
            'output_format': 'local_csv'
        })
        my_scheduler.lock = True
        response = my_scheduler.get_status('my_model_run')

        assert response['status'] == 'done'

    @patch('smif.controller.scheduler.subprocess.Popen')
    def test_status_model_failed(self, mock_popen):
        attrs = {
            'poll.return_value': 1,
            'communicate.return_value': (
                "this is a stdout".encode('utf-8'),
            )
        }
        process_mock = Mock(**attrs)
        mock_popen.return_value = process_mock

        my_scheduler = ModelRunScheduler()
        my_scheduler.add('my_model_run', {
            'directory': 'mock/dir',
            'verbosity': 0,
            'warm_start': False,
            'output_format':
            'local_csv'
        })
        my_scheduler.lock = True
        response = my_scheduler.get_status('my_model_run')

        assert response['status'] == 'failed'

    @patch('smif.controller.scheduler.subprocess.Popen')
    def test_status_model_stopped(self, mock_popen):
        attrs = {
            'poll.return_value': None,
            'communicate.return_value': (
                "this is a stdout".encode('utf-8'),
            )
        }
        process_mock = Mock(**attrs)
        mock_popen.return_value = process_mock

        my_scheduler = ModelRunScheduler()
        my_scheduler.add('my_model_run', {
            'directory': 'mock/dir',
            'verbosity': 0,
            'warm_start': False,
            'output_format':
            'local_csv'
        })
        my_scheduler.lock = True
        my_scheduler.kill('my_model_run')
        response = my_scheduler.get_status('my_model_run')

        assert response['status'] == 'stopped'


class TestJobScheduler():
    @fixture
    def job_graph(self):
        G = networkx.DiGraph()
        a_model = ScenarioModel('a')

        G.add_node(
            'a',
            model=a_model,
            operation=ModelOperation.BEFORE_MODEL_RUN,
            modelrun_name='test',
            current_timestep=1,
            timesteps=[1],
            decision_iteration=0
        )
        b_model = EmptySectorModel('b')
        G.add_node(
            'b',
            model=b_model,
            operation=ModelOperation.SIMULATE,
            modelrun_name='test',
            current_timestep=1,
            timesteps=[1],
            decision_iteration=0
        )
        G.add_edge('a', 'b')
        return G

    @fixture
    def scheduler(self, empty_store):
        empty_store.write_model_run({
            'name': 'test',
            'narratives': {},
            'scenarios': {},
            'sos_model': 'test_sos_model'
        })
        empty_store.write_sos_model({
            'name': 'test_sos_model',
            'scenario_dependencies': [],
            'model_dependencies': []
        })
        scheduler = JobScheduler()
        scheduler.store = empty_store
        return scheduler

    def test_add(self, job_graph, scheduler):
        job_id, err = scheduler.add(job_graph)

        print(err)
        assert err is None
        assert scheduler.get_status(job_id)['status'] == 'done'

    def test_default_status(self):
        scheduler = JobScheduler()
        assert scheduler.get_status(0)['status'] == 'unstarted'

    def test_add_cyclic(self, job_graph, scheduler):
        job_graph.add_edge('b', 'a')
        job_id, err = scheduler.add(job_graph)

        assert isinstance(err, NotImplementedError)
        assert scheduler.get_status(job_id)['status'] == 'failed'

    def test_kill_fails(self, job_graph, scheduler):
        job_id, err = scheduler.add(job_graph)

        assert err is None
        with raises(NotImplementedError):
            scheduler.kill(job_id)

    def test_unknown_operation(self, job_graph, scheduler):
        model = EmptySectorModel('c')

        job_graph.add_node(
            'c',
            model=model,
            operation='unknown_operation',
            modelrun_name='test',
            current_timestep=1,
            timesteps=[1],
            decision_iteration=0
        )
        job_id, err = scheduler.add(job_graph)

        assert isinstance(err, ValueError)
        assert scheduler.get_status(job_id)['status'] == 'failed'



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Water Supply model
- WaterSupplySectorModel implements SectorModel
- wraps ExampleWaterSupplySimulationModel
- instantiate a model instance

"""

import logging

import numpy as np
from smif.model.sector_model import SectorModel


class WaterSupplySectorModel(SectorModel):
    """Example of a class implementing the SectorModel interface,
    using one of the toy water models below to simulate the water supply
    system.
    """
    def simulate(self, data):
        """Simulate water supply

        Arguments
        =========
        data
            - inputs/parameters, implicitly includes:
                - scenario data, e.g. expected level of rainfall
                - data output from other models in workflow
                - parameters set or adjusted for this model run
                - system state data, e.g. reservoir level at year start
            - system state, implicity includes:
                - initial existing system/network
                - decisions, e.g. asset build instructions, demand-side interventions to apply
        """
        # State

        current_interventions = data.get_current_interventions()

        self.logger.debug("Current interventions: {}".format(current_interventions))
        number_of_treatment_plants = 2

        # Inputs
        per_capita_water_demand = data.get_parameter(
            'per_capita_water_demand').as_ndarray()  # liter/person
        population = data.get_data('population').as_ndarray()  # people

        water_demand = data.get_data('water_demand').as_ndarray()  # liter
        final_water_demand = (population * per_capita_water_demand) + water_demand

        raininess = sum(data.get_data('precipitation').as_ndarray())  # milliliters to mega
        if data.current_timestep == data.base_timestep:
            reservoir_level = data.get_data('reservoir_level', 2009)
        else:
            reservoir_level = data.get_previous_timestep_data('reservoir_level')

        reservoir_level = sum(reservoir_level.as_ndarray())  # megaliters
        self.logger.info('Total reservoir level before timestep: %s', reservoir_level)

        self.logger.debug(
            "Parameters:\n "
            "Population: %s\n"
            "Raininess: %s\n "
            "Reservoir level: %s\n  "
            "Final demand: %s\n",
            population.sum(),
            raininess.sum(),
            reservoir_level,
            final_water_demand
        )

        # Parameters
        self.logger.debug(data.get_parameters())

        # simulate (wrapping toy model)
        instance = ExampleWaterSupplySimulationModel()
        instance.raininess = raininess
        instance.number_of_treatment_plants = number_of_treatment_plants
        instance.reservoir_level = reservoir_level

        # run
        water, cost = instance.run()

        self.logger.info(
            "Water: %s, Cost: %s, Reservoir: %s", water, cost, instance.reservoir_level)

        # set results
        data.set_results('water', np.ones((3, )) * water / 3)
        data.set_results("cost", np.ones((3, )) * cost / 3)
        data.set_results("energy_demand", np.ones((3, )) * 3)
        data.set_results("reservoir_level", np.ones((3, )) * instance.reservoir_level / 3)

    def extract_obj(self, results):
        return results['cost'].sum()


class ExampleWaterSupplySimulationModel(object):
    """An example simulation model used for testing purposes

    Parameters
    ==========
    raininess : int
        The amount of rain produced in each simulation
    number_of_treatment_plants : int
        The amount of water is a function of the number of treatment plants and
        the amount of raininess
    """
    def __init__(self,
                 raininess=None,
                 number_of_treatment_plants=None,
                 reservoir_level=None):
        self.raininess = raininess
        self.number_of_treatment_plants = number_of_treatment_plants
        self.reservoir_level = reservoir_level

    def run(self):
        """Runs the water supply model

        Only 1 unit of water is produced per treatment plant,
        no matter how rainy.

        Each treatment plant costs 1.0 unit.
        """
        logger = logging.getLogger(__name__)

        logger.debug("There are %s plants", self.number_of_treatment_plants)
        logger.debug("It is %s rainy", self.raininess)

        logger.debug("Reservoir level was %s", self.reservoir_level)
        self.reservoir_level += self.raininess

        water = min(self.number_of_treatment_plants, self.reservoir_level)
        logger.debug("The system produces %s water", water)

        self.reservoir_level -= water
        logger.debug("Reservoir level now %s", self.reservoir_level)

        cost = 1.264 * self.number_of_treatment_plants
        logger.debug("The system costs £%s", cost)

        return water, cost


if __name__ == '__main__':
    """Run core model if this script is run from the command line
    """
    CORE_MODEL = ExampleWaterSupplySimulationModel(1, 1, 2)
    CORE_MODEL.run()



"""Handles conversion between units used in the `SosModel`
"""
from pint import DimensionalityError, UndefinedUnitError, UnitRegistry  # type: ignore

from smif.convert.adaptor import Adaptor
from smif.data_layer.data_handle import DataHandle


class UnitAdaptor(Adaptor):
    """Scalar conversion of units
    """
    def __init__(self, name):
        self._register = UnitRegistry()
        super().__init__(name)

    def before_model_run(self, data_handle: DataHandle):
        """Register unit definitions in registry before model run
        """
        units = data_handle.read_unit_definitions()
        for unit in units:
            self._register.define(unit)

    def convert(self, data_array, to_spec, coefficients):
        data = data_array.data
        from_spec = data_array.spec

        try:
            quantity = self._register.Quantity(data, from_spec.unit)
        except UndefinedUnitError:
            raise ValueError('Cannot convert from undefined unit {}'.format(from_spec.unit))

        try:
            converted_quantity = quantity.to(to_spec.unit)
        except UndefinedUnitError as ex:
            raise ValueError('Cannot convert undefined unit {}'.format(to_spec.unit)) from ex
        except DimensionalityError as ex:
            msg = 'Cannot convert unit from {} to {}'
            raise ValueError(msg.format(from_spec.unit, to_spec.unit)) from ex

        return converted_quantity.magnitude

    def get_coefficients(self, data_handle, from_spec, to_spec):
        # override with no-op - all the work is done in convert with scalar operations
        pass

    def generate_coefficients(self, from_spec, to_spec):
        # override with no-op - all the work is done in convert with scalar operations
        pass

    def parse_unit(self, unit_string):
        """Parse a unit string (abbreviation or full) into a Unit object

        Parameters
        ----------
        unit : str

        Returns
        -------
        quantity : :class:`pint.Unit`
        """
        try:
            unit = self._register.parse_units(unit_string)
        except UndefinedUnitError:
            self.logger.warning("Unrecognised unit: %s", unit_string)
            unit = None
        return unit



"""Test DataArray
"""
# pylint: disable=redefined-outer-name
import numpy
import pandas as pd
import xarray as xr
from numpy.testing import assert_array_equal
from pytest import fixture, raises
from smif.data_layer.data_array import DataArray, show_null
from smif.exception import SmifDataNotFoundError, SmifDataMismatchError
from smif.metadata import Spec


@fixture
def dims():
    return ['a', 'b', 'c']


@fixture
def coords():
    return [['a1', 'a2'], ['b1', 'b2', 'b3'], ['c1', 'c2', 'c3', 'c4']]


@fixture
def spec(dims, coords):
    return Spec(
        name='test_data',
        dims=dims,
        coords={
            'a': coords[0],
            'b': coords[1],
            'c': coords[2],
        },
        dtype='float',
        abs_range=(0, 1),
        exp_range=(0, 0.5)
    )


@fixture
def data():
    return numpy.arange(24, dtype='float').reshape((2, 3, 4))


@fixture
def non_numeric_spec(dims, coords):
    return Spec(
        name='test_data',
        dims=dims,
        coords={
            'a': coords[0],
            'b': coords[1],
            'c': coords[2],
        },
        dtype='str',
        abs_range=('0', '1'),
        exp_range=('0', '0.5')
    )


@fixture
def non_numeric_data():
    strings = [str(x) for x in range(24)]
    return numpy.array(strings, dtype=numpy.object).reshape((2, 3, 4))


@fixture
def small_da(spec, data):
    return DataArray(spec, data)


@fixture
def small_da_non_numeric(non_numeric_spec, non_numeric_data):
    return DataArray(non_numeric_spec, non_numeric_data)


@fixture
def small_da_df(spec, data):
    index = pd.MultiIndex.from_product([c.ids for c in spec.coords], names=spec.dims)
    return pd.DataFrame({spec.name: numpy.reshape(data, data.size)}, index=index)


@fixture
def small_da_xr(spec, data):
    return xr.DataArray(data, [(c.name, c.ids) for c in spec.coords])


class TestDataArray():
    def test_init(self, spec, data):
        """Should initialise from spec and ndarray of data
        """
        da = DataArray(spec, data)
        numpy.testing.assert_equal(da.data, data)
        assert spec == da.spec

    def test_rename(self, small_da):
        """Allow setting a Spec name
        """
        assert small_da.name == 'test_data'
        small_da.name = 'test'
        assert small_da.name == 'test'

    def test_as_df(self, small_da, small_da_df):
        """Should create a pandas.DataFrame from a DataArray
        """
        actual = small_da.as_df()
        pd.testing.assert_frame_equal(actual, small_da_df)

    def test_from_df(self, small_da, small_da_df):
        """Should create a DataArray from a pandas.DataFrame
        """
        actual = DataArray.from_df(small_da.spec, small_da_df)
        assert actual == small_da

    def test_from_df_partial(self, spec):
        """Should create a DataArray that can handle missing data, returning nan/null
        """
        df = pd.DataFrame({
            'a': ['a1'],
            'b': ['b1'],
            'c': ['c2'],
            'test_data': [1]
        }).set_index(['a', 'b', 'c'])
        expected_data = numpy.full(spec.shape, numpy.nan)
        expected_data[0, 0, 1] = 1.0
        expected = DataArray(spec, expected_data)

        actual = DataArray.from_df(spec, df)

        assert_array_equal(actual.data, expected.data)
        assert actual == expected

    def test_combine(self, small_da, data):
        """Should override values where present (use case: full array of default values,
        overridden by a partial array of specific values).

        See variously:
        - http://xarray.pydata.org/en/stable/combining.html#merging-with-no-conflicts
        - https://pandas.pydata.org/pandas-docs/stable/generated/pandas.DataFrame.update.html
        """
        partial_data = numpy.full(small_da.shape, numpy.nan)
        partial_data[0, 0, 1] = 99
        partial = DataArray(small_da.spec, partial_data)

        # update in-place
        small_da.update(partial)

        # match fixture data
        expected_data = numpy.arange(24, dtype='float').reshape((2, 3, 4))
        expected_data[0, 0, 1] = 99
        expected = DataArray(small_da.spec, expected_data)

        assert small_da == expected
        assert_array_equal(small_da.data, expected.data)

    def test_as_xarray(self, small_da, small_da_xr):
        actual = small_da.as_xarray()
        xr.testing.assert_equal(actual, small_da_xr)

    def test_from_xarray(self, small_da, small_da_xr):
        actual = DataArray.from_xarray(small_da.spec, small_da_xr)
        assert actual == small_da

    def test_as_ndarray(self, small_da, data):
        actual = small_da.as_ndarray()
        expected = data
        assert_array_equal(actual, expected)

    def test_equality(self, small_da, spec, data):
        expected = DataArray(spec, data)
        assert small_da == expected

    def test_repr(self, small_da, spec, data):
        assert repr(small_da) == "<DataArray('{}', '{}')>".format(spec, data)

    def test_dim_coords(self, small_da, spec):
        actual = small_da.dim_coords('a')
        expected = spec.dim_coords('a')
        assert actual == expected

    def test_dim_names(self, small_da, spec):
        actual = small_da.dim_names('a')
        expected = spec.dim_names('a')
        assert actual == expected

    def test_dim_elements(self, small_da, spec):
        actual = small_da.dim_elements('a')
        expected = spec.dim_elements('a')
        assert actual == expected

    def test_coords(self, small_da, spec):
        actual = small_da.coords
        expected = spec.coords
        assert actual == expected


class TestDataFrameInterop():
    def test_to_from_df(self):
        df = pd.DataFrame([
            {
                'test': 3,
                'region': 'oxford',
                'interval': 1
            }
        ]).set_index(['region', 'interval'])

        spec = Spec(
            name='test',
            dims=['region', 'interval'],
            coords={'region': ['oxford'], 'interval': [1]},
            dtype='int64'
        )

        da = DataArray(spec, numpy.array([[3.]], dtype='int64'))
        da_from_df = DataArray.from_df(spec, df)
        assert da_from_df == da

        da_to_df = da.as_df()
        pd.testing.assert_frame_equal(da_to_df, df)

    def test_multi_dim_order(self):
        spec = Spec(
            name='test',
            coords={'lad': ['c', 'a', 'b'], 'interval': [4, 2]},
            dims=['lad', 'interval'],
            dtype='float'
        )
        data = numpy.array([
            #4  2
            [1, 2],  # c
            [5, 6],  # a
            [9, 0]   # b
        ], dtype='float')
        da = DataArray(spec, data)

        df = pd.DataFrame([
            {'test': 6.0, 'lad': 'a', 'interval': 2},
            {'test': 0.0, 'lad': 'b', 'interval': 2},
            {'test': 2.0, 'lad': 'c', 'interval': 2},
            {'test': 5.0, 'lad': 'a', 'interval': 4},
            {'test': 9.0, 'lad': 'b', 'interval': 4},
            {'test': 1.0, 'lad': 'c', 'interval': 4},
        ]).set_index(['lad', 'interval'])
        da_from_df = DataArray.from_df(spec, df)
        assert da_from_df == da

        da_to_df = da.as_df().sort_index()
        df = df.sort_index()
        pd.testing.assert_frame_equal(da_to_df, df)

    def test_match_metadata(self):
        spec = Spec(
            name='test',
            dims=['region'],
            coords={'region': ['oxford']},
            dtype='int64'
        )

        # must have a column named the same as the spec.name
        df = pd.DataFrame([
            {'region': 'oxford', 'other': 'else'}
        ]).set_index(['region'])
        msg = "Data for 'test' expected a data column called 'test' and index names " + \
              "['region'], instead got data columns ['other'] and index names ['region']"
        with raises(SmifDataMismatchError) as ex:
            DataArray.from_df(spec, df)
        assert msg in str(ex)

        # may not be indexed, if columns are otherwise all okay
        df = pd.DataFrame([
            {'region': 'oxford', 'test': 1}
        ])
        DataArray.from_df(spec, df)

        # must have an index level for each spec dimension
        df = pd.DataFrame([
            {'test': 3.14}
        ])
        msg = "Data for 'test' expected a data column called 'test' and index names " + \
              "['region'], instead got data columns ['test'] and index names [None]"
        with raises(SmifDataMismatchError) as ex:
            DataArray.from_df(spec, df)
        assert msg in str(ex)

        # must not have dimension labels outside of the spec dimension
        df = pd.DataFrame([
            {'test': 3.14, 'region': 'oxford'},
            {'test': 3.14, 'region': 'extra'}
        ]).set_index(['region'])
        msg = "Data for 'test' contained unexpected values in the set of coordinates for " + \
              "dimension 'region': ['extra']"
        with raises(SmifDataMismatchError) as ex:
            DataArray.from_df(spec, df)
        assert msg in str(ex)

    def test_scalar(self):
        # should handle zero-dimensional case (numpy array as scalar)
        data = numpy.array(2.0)
        spec = Spec(
            name='test',
            dims=[],
            coords={},
            dtype='float'
        )
        da = DataArray(spec, data)
        df = pd.DataFrame([{'test': 2.0}])
        da_from_df = DataArray.from_df(spec, df)
        assert da_from_df == da

        df_from_da = da.as_df()
        pd.testing.assert_frame_equal(df_from_da, df)

    def test_error_duplicate_rows_single_index(self):
        spec = Spec(
            name='test',
            dims=['a'],
            coords={'a': [1, 2]},
            dtype='int'
        )
        df = pd.DataFrame([
            {'a': 1, 'test': 0},
            {'a': 2, 'test': 1},
            {'a': 1, 'test': 2},
        ])

        with raises(SmifDataMismatchError) as ex:
            DataArray.from_df(spec, df)

        msg = "Data for 'test' contains duplicate values at [{'a': 1}]"
        assert msg in str(ex)

    def test_error_duplicate_rows_multi_index(self):
        spec = Spec(
            name='test',
            dims=['a', 'b'],
            coords={'a': [1, 2], 'b': [3, 4]},
            dtype='int'
        )
        df = pd.DataFrame([
            {'a': 1, 'b': 3, 'test': 0},
            {'a': 2, 'b': 3, 'test': 1},
            {'a': 1, 'b': 4, 'test': 2},
            {'a': 2, 'b': 4, 'test': 3},
            {'a': 2, 'b': 4, 'test': 4},
        ])

        with raises(SmifDataMismatchError) as ex:
            DataArray.from_df(spec, df)

        msg = "Data for 'test' contains duplicate values at [{'a': 2, 'b': 4}]"
        msg_alt = "Data for 'test' contains duplicate values at [{'b': 4, 'a': 2}]"
        assert msg in str(ex) or msg_alt in str(ex)


class TestMissingData:

    def test_missing_data_raises(self, small_da):
        """Should check for NaNs and raise SmifDataError
        """
        da = small_da
        da.validate_as_full()
        da.data[1, 1] = numpy.NaN

        with raises(SmifDataMismatchError) as ex:
            da.validate_as_full()

        msg = "Data for 'test_data' had missing values - read 20 but expected 24 in " + \
              "total, from dims of length {a: 2, b: 3, c: 4}"
        assert msg in str(ex)

    def test_missing_data_message(self, small_da):
        """Should check for NaNs and raise SmifDataError
        """
        da = small_da
        da.validate_as_full()
        da.data[1, 1, 1] = numpy.nan
        da.data[0, 0, 3] = numpy.nan
        with raises(SmifDataMismatchError) as ex:
            da.validate_as_full()

        expected = "Data for 'test_data' had missing values - read 22 but expected 24 in " + \
                   "total, from dims of length {a: 2, b: 3, c: 4}"
        assert expected in str(ex)

    def test_missing_data_message_non_numeric(self, small_da_non_numeric):
        """Should check for NaNs and raise SmifDataError
        """
        da = small_da_non_numeric
        da.validate_as_full()
        da.data[1, 1, 1] = None
        da.data[0, 0, 3] = None
        with raises(SmifDataMismatchError) as ex:
            da.validate_as_full()

        expected = "Data for 'test_data' had missing values - read 22 but expected 24 in " + \
                   "total, from dims of length {a: 2, b: 3, c: 4}"
        assert expected in str(ex)

    def test_no_missing_data(self, small_da):

        df = small_da.as_df()
        actual = show_null(df)
        expected = pd.DataFrame(columns=['test_data'], dtype=float)
        levels = [['a1', 'a2'], ['b1', 'b2', 'b3'], ['c1', 'c2', 'c3', 'c4']]
        codes = [[], [], []]
        names = ['a', 'b', 'c']
        try:
            expected.index = pd.MultiIndex(levels=levels, codes=codes, names=names)
        except TypeError:
            expected.index = pd.MultiIndex(levels=levels, labels=codes, names=names)

        pd.testing.assert_frame_equal(actual, expected)

    def test_no_missing_data_non_numeric(self, small_da_non_numeric):

        df = small_da_non_numeric.as_df()
        actual = show_null(df)
        expected = pd.DataFrame(columns=['test_data'], dtype=str)
        levels = [['a1', 'a2'], ['b1', 'b2', 'b3'], ['c1', 'c2', 'c3', 'c4']]
        codes = [[], [], []]
        names = ['a', 'b', 'c']
        try:
            expected.index = pd.MultiIndex(levels=levels, codes=codes, names=names)
        except TypeError:
            expected.index = pd.MultiIndex(levels=levels, labels=codes, names=names)

        pd.testing.assert_frame_equal(actual, expected)

    def test_missing_data_non_numeric(self, small_da_non_numeric):

        small_da_non_numeric.data[1, 1, 1] = None
        df = small_da_non_numeric.as_df()
        actual = show_null(df)
        levels = [['a1', 'a2'], ['b1', 'b2', 'b3'], ['c1', 'c2', 'c3', 'c4']]
        codes = [[1], [1], [1]]
        names = ['a', 'b', 'c']
        try:
            index = pd.MultiIndex(levels=levels, codes=codes, names=names)
        except TypeError:
            index = pd.MultiIndex(levels=levels, labels=codes, names=names)


        expected = pd.DataFrame(data=numpy.array([[None]], dtype=numpy.object),
                                index=index,
                                columns=['test_data'])
        pd.testing.assert_frame_equal(actual, expected)



"""File-backed metadata store
"""
import copy
import json
import os
from functools import lru_cache
from logging import getLogger
from typing import Dict, List, Union

import pandas  # type: ignore
from pandas import compat as pandas_compat
from pandas.core import common as pandas_common  # type: ignore
from ruamel.yaml import YAML  # type: ignore
from smif.data_layer.abstract_metadata_store import MetadataStore
from smif.exception import SmifDataNotFoundError, SmifDataReadError

# Import fiona if available (optional dependency)
try:
    import fiona  # type: ignore
except ImportError:
    pass


class FileMetadataStore(MetadataStore):
    """File-based metadata store (supports YAML, CSV, or GDAL-compatible files)
    """
    def __init__(self, base_folder):
        super().__init__()
        self.logger = getLogger(__name__)

        base_folder = str(base_folder)
        self.units_path = os.path.join(base_folder, 'data', 'user-defined-units.txt')
        self.data_folder = os.path.join(base_folder, 'data', 'dimensions')
        self.config_folder = os.path.join(base_folder, 'config', 'dimensions')

    # region Units
    def read_unit_definitions(self) -> List[str]:
        try:
            with open(self.units_path, 'r') as units_fh:
                return [line.strip() for line in units_fh]
        except FileNotFoundError:
            self.logger.warning('Units file not found, expected at %s', str(self.units_path))
            return []

    def write_unit_definitions(self, definitions: List[str]):
        with open(self.units_path, 'w') as units_fh:
            units_fh.writelines(definitions)
    # endregion

    # region Dimensions
    def read_dimensions(self, skip_coords=False) -> List[dict]:
        dim_names = _read_filenames_in_dir(self.config_folder, '.yml')
        return [self.read_dimension(name, skip_coords) for name in dim_names]

    def read_dimension(self, dimension_name: str, skip_coords=False):
        dim = _read_yaml_file(self.config_folder, dimension_name)
        if skip_coords:
            del dim['elements']
        else:
            dim['elements'] = self._read_dimension_file(dim['elements'])
        return dim

    def write_dimension(self, dimension: Dict):
        # write elements to csv file (by default, can handle any nested data)
        elements_filename = "{}.csv".format(dimension['name'])
        elements = dimension['elements']
        self._write_dimension_file(elements_filename, elements)

        # refer to elements by filename and add to config
        dimension_with_ref = copy.copy(dimension)
        dimension_with_ref['elements'] = elements_filename
        _write_yaml_file(self.config_folder, dimension['name'], dimension_with_ref)

    def update_dimension(self, dimension_name: str, dimension: Dict):
        # look up elements filename and write elements

        old_dim = _read_yaml_file(self.config_folder, dimension_name)
        elements_filename = old_dim['elements']
        elements = dimension['elements']
        self._write_dimension_file(elements_filename, elements)

        # refer to elements by filename and update config
        dimension_with_ref = copy.copy(dimension)
        dimension_with_ref['elements'] = elements_filename

        _write_yaml_file(self.config_folder, dimension_name, dimension_with_ref)

    def delete_dimension(self, dimension_name: str):
        # read to find filename

        old_dim = _read_yaml_file(self.config_folder, dimension_name)
        elements_filename = old_dim['elements']
        # remove elements data
        os.remove(os.path.join(self.data_folder, elements_filename))
        # remove description
        os.remove(os.path.join(self.config_folder, "{}.yml".format(dimension_name)))

    @lru_cache(maxsize=32)
    def _read_dimension_file(self, filename: str) -> List[Dict]:
        filepath = os.path.join(self.data_folder, filename)
        filebasename, ext = os.path.splitext(filename)
        if ext == '.csv':
            dataframe = pandas.read_csv(filepath)
            data = _df_to_records(dataframe)
            if 'interval' in data[0]:
                data = self._unstringify_interval(data)
        elif ext in ('.geojson', '.shp'):
            data = self._read_spatial_file(filepath)
        else:
            msg = "Extension '{}' not recognised, expected one of ('.csv', "
            msg += "'.geojson', '.shp') when reading {}"
            raise SmifDataReadError(msg.format(ext, filepath))
        return data

    def _write_dimension_file(self, filename: str, data: List[Dict]):
        # lru_cache may now be invalid, so clear it
        self._read_dimension_file.cache_clear()
        path = os.path.join(self.data_folder, filename)
        filebasename, ext = os.path.splitext(filename)
        if ext == '.csv':
            if 'interval' in data[0]:
                data = self._stringify_interval(data)
            pandas.DataFrame.from_records(data).to_csv(path, index=False)
        elif ext in ('.geojson', '.shp'):
            raise NotImplementedError("Writing spatial dimensions not yet supported")
            # self._write_spatial_file(filepath)
        else:
            msg = "Extension '{}' not recognised, expected one of ('.csv', "
            msg += "'.geojson', '.shp') when writing {}"
            raise SmifDataReadError(msg.format(ext, path))
        return data

    def _stringify_interval(self, data: List[Dict]) -> List[Dict]:
        output = []
        for item in data:
            output_item = copy.copy(item)
            try:
                output_item['interval'] = json.dumps(item['interval'])
            except KeyError:
                self.logger.warning("Expected interval in element %s", item)
            output.append(output_item)
        return output

    def _unstringify_interval(self, data: List[Dict]) -> List[Dict]:
        output = []
        for item in data:
            output_item = copy.copy(item)
            try:
                output_item['interval'] = json.loads(item['interval'])
            except KeyError:
                self.logger.warning("Expected interval in element %s", item)
            output.append(output_item)
        return output
    # endregion

    @staticmethod
    def _read_spatial_file(filepath) -> List[Dict]:
        try:
            with fiona.drivers():
                with fiona.open(filepath) as src:
                    data = []
                    for feature in src:
                        element = {
                            'name': feature['properties']['name'],
                            'feature': feature
                        }
                        data.append(element)
            return data
        except NameError as ex:
            msg = "Could not read spatial dimension definition '%s' " % (filepath)
            msg += "Please install fiona to read geographic data files. Try running: \n"
            msg += "    pip install smif[spatial]\n"
            msg += "or:\n"
            msg += "    conda install fiona shapely rtree\n"
            raise SmifDataReadError(msg) from ex
        except IOError as ex:
            msg = "Could not read spatial dimension definition '%s' " % (filepath)
            msg += "Please verify that the path is correct and "
            msg += "that the file is present on this location."
            raise SmifDataNotFoundError(msg) from ex


def _read_yaml_file(directory, name):
    """Parse yaml config file into plain data (lists, dicts and simple values)

    Parameters
    ----------
    directory : str
    name : str
        file basename (without yml extension)
    """
    path = os.path.join(directory, "{}.yml".format(name))
    with open(path, 'r') as file_handle:
        return YAML().load(file_handle)


def _write_yaml_file(directory, name, data):
    """Write plain data to a file as yaml

    Parameters
    ----------
    directory : str
    name : str
        file basename (without yml extension)
    data
        Data to write (should be lists, dicts and simple values)
    """
    path = os.path.join(directory, "{}.yml".format(name))
    with open(path, 'w') as file_handle:
        yaml = YAML()
        yaml.default_flow_style = False
        yaml.allow_unicode = True
        return yaml.dump(data, file_handle)


def _read_filenames_in_dir(path, extension):
    """Returns the name of the Yaml files in a certain directory

    Arguments
    ---------
    path: str
        Path to directory
    extension: str
        Extension of files (such as: '.yml' or '.csv')

    Returns
    -------
    list
        The list of files in `path` with extension
    """
    files = []
    for filename in os.listdir(path):
        if filename.endswith(extension):
            basename, _ = os.path.splitext(filename)
            files.append(basename)
    return files


def _df_to_records(dataframe: pandas.DataFrame) -> List[Dict]:
    """Fix pandas conversion to list[dict] with python scalar values

    Ported here from future release of pandas 0.24.0

    See:
    - PR: https://github.com/pandas-dev/pandas/pull/23921
    - Issue: https://github.com/pandas-dev/pandas/issues/23753

    Note that this skips the pandas_common,maybe_box_datetimelike implementation, which may be
    desired but relies on more pandas internals so is not copied over (yet).
    """
    into_c = pandas_common.standardize_mapping(dict)
    return [
        into_c(
            (k, v)
            for k, v in pandas_compat.iteritems(row._asdict())
        )
        for row in dataframe.itertuples(index=False)
    ]



"""DataArray provides a thin wrapper around multidimensional arrays and metadata
"""
from logging import getLogger

import numpy as np  # type: ignore
from smif.exception import (SmifDataError, SmifDataMismatchError,
                            SmifDataNotFoundError)
from smif.metadata.spec import Spec

# Import pandas, xarray if available (optional dependencies)
try:
    import pandas  # type: ignore
    import xarray  # type: ignore
except ImportError:
    pass


INSTALL_WARNING = """\
Please install pandas and xarray to access smif.DataArray
data as pandas.DataFrame or xarray.DataArray. Try running:
    pip install smif[data]
or:
    conda install pandas xarray
"""


class DataArray():
    """DataArray provides access to input/parameter/results data, with conversions to common
    python data libraries (for example: numpy, pandas, xarray).

    Attributes
    ----------
    spec: smif.metadata.spec.Spec
    data: numpy.ndarray
    """
    def __init__(self, spec: Spec, data: np.ndarray):
        self.logger = getLogger(__name__)

        if not hasattr(data, 'shape'):
            self.logger.debug("Data is not an numpy.ndarray")
            data = np.array(data)

        if not hasattr(spec, 'shape'):
            self.logger.error("spec argument is not a Spec")
            raise TypeError("spec argument is not a Spec")

        if not data.shape == spec.shape:
            # special case for scalar - allow a single-value 1D array, here coerced to single
            # value 0D array. Then simpler to create from DataFrame or xarray.DataArray
            if data.shape == (1,) and spec.shape == ():
                data = np.array(data[0])
            else:
                msg = "Data shape {} does not match spec {}"
                raise SmifDataMismatchError(msg.format(data.shape, spec.shape))

        self.spec = spec
        self.data = data

    def __eq__(self, other):
        return self.spec == other.spec and \
            _array_equal_nan(self.data, other.data)

    def __repr__(self):
        return "<DataArray('{}', '{}')>".format(self.spec, self.data)

    def __str__(self):
        return "<DataArray('{}', '{}')>".format(self.spec, self.data)

    def as_dict(self):
        """
        """
        return self.spec.as_dict()

    @property
    def name(self):
        """The name of the data that this spec describes.
        """
        return self.spec.name

    @name.setter
    def name(self, value):
        self.spec.name = value

    @property
    def description(self):
        """A human-friendly description
        """
        return self.spec.description

    @property
    def dims(self):
        """Names for each dimension
        """
        return self.spec.dims

    @property
    def coords(self):
        """Coordinate labels for each dimension.
        """
        return self.spec.coords

    def dim_coords(self, dim):
        """Coordinates for a given dimension
        """
        return self.spec.dim_coords(dim)

    def dim_names(self, dim):
        """Coordinate names for a given dimension
        """
        return self.spec.dim_names(dim)

    def dim_elements(self, dim):
        """Coordinate elements for a given dimension
        """
        return self.spec.dim_elements(dim)

    @property
    def unit(self):
        """The unit for all data points.
        """
        return self.spec.unit

    @property
    def shape(self):
        """The shape of the data array
        """
        return self.data.shape

    def as_ndarray(self) -> np.ndarray:
        """Access as a :class:`numpy.ndarray`
        """
        return self.data

    def as_df(self) -> pandas.DataFrame:
        """Access DataArray as a :class:`pandas.DataFrame`
        """
        dims = self.dims
        coords = [c.ids for c in self.coords]

        try:
            if dims and coords:
                index = pandas.MultiIndex.from_product(coords, names=dims)
                return pandas.DataFrame(
                    {self.name: np.reshape(self.data, self.data.size)}, index=index)
            else:
                # with no dims or coords, should be in the zero-dimensional case
                if self.data.shape != ():
                    msg = "Expected zero-dimensional data, got %s" % self.data.shape
                    raise SmifDataMismatchError(msg)
                return pandas.DataFrame([{self.name: self.data[()]}])
        except NameError as ex:
            raise SmifDataError(INSTALL_WARNING) from ex

    @classmethod
    def from_df(cls, spec, dataframe):
        """Create a DataArray from a :class:`pandas.DataFrame`
        """
        name = spec.name
        dims = spec.dims

        data_columns = dataframe.columns.values.tolist()
        index_names = dataframe.index.names

        if dims and len(index_names) == 1 and index_names[0] is None:
            # case when an unindexed dataframe was passed in, try to recover automagically
            if set(dims).issubset(set(data_columns)):
                dataframe = dataframe.set_index(dims)
                data_columns = dataframe.columns.values.tolist()
                index_names = dataframe.index.names

        if name not in data_columns or (dims and set(dims) != set(index_names)):
            msg = "Data for '{name}' expected a data column called '{name}' and index " + \
                  "names {dims}, instead got data columns {data_columns} and index names " + \
                  "{index_names}"
            raise SmifDataMismatchError(msg.format(
                name=name,
                dims=dims,
                data_columns=data_columns,
                index_names=index_names))

        try:
            # convert to dataset
            xr_dataset = dataframe.to_xarray()

            # extract xr.DataArray
            xr_data_array = xr_dataset[spec.name]

            # reindex to ensure data order and fill out NaNs
            xr_data_array = _reindex_xr_data_array(spec, xr_data_array)

        # xarray raises Exception in v0.10 (narrowed to ValueError in v0.11)
        except Exception as ex:  # pylint: disable=broad-except
            dups = find_duplicate_indices(dataframe)
            if dups:
                msg = "Data for '{name}' contains duplicate values at {dups}"
                raise SmifDataMismatchError(msg.format(name=name, dups=dups)) from ex
            else:
                raise ex

        return cls(spec, xr_data_array.data)

    def as_xarray(self):
        """Access DataArray as a :class:`xarray.DataArray`
        """
        metadata = self.spec.as_dict()
        del metadata['dims']
        del metadata['coords']

        dims = self.dims
        coords = {c.name: c.ids for c in self.coords}

        try:
            return xarray.DataArray(
                self.data,
                coords=coords,
                dims=dims,
                name=self.name,
                attrs=metadata
            )
        except NameError as ex:
            raise SmifDataError(INSTALL_WARNING) from ex

    @classmethod
    def from_xarray(cls, spec, xr_data_array):
        """Create a DataArray from a :class:`xarray.DataArray`
        """
        # reindex to ensure data order and fill out NaNs
        xr_data_array = _reindex_xr_data_array(spec, xr_data_array)
        return cls(spec, xr_data_array.data)

    def update(self, other):
        """Update data values with any from other which are non-null
        """
        assert self.spec == other.spec, "Specs must match when updating DataArray"
        # convert self and other to xarray representation
        self_xr = self.as_xarray()
        other_xr = other.as_xarray()
        # use xarray.combine_first convenience function
        overridden = other_xr.combine_first(self_xr)
        # assign result back to self
        self.data = overridden.data

    def validate_as_full(self):
        """Check that the data array contains no NaN values
        """
        dataframe = self.as_df()
        if np.any(dataframe.isnull()):
            expected_len = len(dataframe)
            missing_data = show_null(dataframe)
            actual_len = expected_len - len(missing_data)
            dim_lens = "{" + ", ".join(
                "{}: {}".format(dim, len_) for dim, len_ in zip(self.dims, self.shape)
            ) + "}"
            self.logger.debug("Missing data:\n\n    %s", missing_data)
            msg = "Data for '{name}' had missing values - read {actual_len} but expected " + \
                  "{expected_len} in total, from dims of length {dim_lens}"
            raise SmifDataMismatchError(msg.format(
                name=self.name,
                actual_len=actual_len,
                expected_len=expected_len,
                dim_lens=dim_lens))


def show_null(dataframe) -> pandas.DataFrame:
    """Shows missing data

    Returns
    -------
    pandas.DataFrame
    """
    try:
        missing_data = dataframe[dataframe.isnull().values]
    except NameError as ex:
        raise SmifDataError(INSTALL_WARNING) from ex
    return missing_data


def find_duplicate_indices(dataframe):
    """Find duplicate indices in a DataFrame

    Returns
    -------
    list[dict]
    """
    # find duplicate index entries
    dups_df = dataframe[dataframe.index.duplicated()]
    # drop data columns, reset index to promote index to values
    dups_index_df = dups_df.drop(dups_df.columns, axis=1).reset_index()
    return dups_index_df.to_dict('records')


def _array_equal_nan(a, b):
    """Compare numpy arrays for equality, allowing NaN to be considerd equal to itself
    """
    if np.issubdtype(a.dtype, np.number) and np.issubdtype(b.dtype, np.number):
        return np.all((a == b) | (np.isnan(a) & np.isnan(b)))
    else:
        return np.all(a == b)


def _reindex_xr_data_array(spec, xr_data_array):
    """Reindex to ensure full data, order
    """
    # all index values must exist in dimension - extras would otherwise be silently dropped
    for dim in spec.dims:
        index_values = set(xr_data_array.coords[dim].values)
        dim_names = set(spec.dim_names(dim))
        in_index_but_not_dim_names = index_values - dim_names
        if in_index_but_not_dim_names:
            msg = "Data for '{name}' contained unexpected values in the set of " + \
                  "coordinates for dimension '{dim}': {extras}"
            raise SmifDataMismatchError(msg.format(
                dim=dim, extras=list(in_index_but_not_dim_names), name=spec.name))

    coords = {c.name: c.ids for c in spec.coords}
    xr_data_array = xr_data_array.reindex(indexers=coords)

    return xr_data_array



"""Test aggregation/disaggregation of data between sets of areas
"""
from copy import copy
from unittest.mock import Mock

import numpy as np
from pytest import fixture, raises
from smif.convert.region import RegionAdaptor, RegionSet
from smif.convert.register import NDimensionalRegister
from smif.data_layer.data_array import DataArray
from smif.metadata import Spec


@fixture(scope='function')
def register(regions_half_triangles, regions_half_squares, regions_single_half_square,
             regions_rect):
    """Region register with regions pre-registered
    """
    register = NDimensionalRegister()
    register.register(RegionSet('half_triangles', regions_half_triangles))
    register.register(RegionSet('half_squares', regions_half_squares))
    register.register(RegionSet('single_half_square', regions_single_half_square))
    register.register(RegionSet('rect', regions_rect))
    alt = copy(regions_rect)
    register.register(RegionSet('rect_alt', alt))
    return register


class TestRegionAdaptor:
    """Converting between region sets, assuming uniform distribution as necessary
    """
    def test_aggregate_region(self, regions_rect, regions_half_squares):
        """Two regions aggregated to one, one interval
        """
        adaptor = RegionAdaptor('test-square-half')
        from_spec = Spec(
            name='test-var',
            dtype='float',
            dims=['half_squares'],
            coords={
                'half_squares': regions_half_squares
            }
        )
        adaptor.add_input(from_spec)
        to_spec = Spec(
            name='test-var',
            dtype='float',
            dims=['rect'],
            coords={
                'rect': regions_rect
            }
        )
        adaptor.add_output(to_spec)

        actual_coefficients = adaptor.generate_coefficients(from_spec, to_spec)
        expected = np.ones((2, 1))  # aggregating coefficients
        np.testing.assert_allclose(actual_coefficients, expected, rtol=1e-3)

        data = np.array([24, 24])  # area a,b

        data_array = DataArray(from_spec, data)

        data_handle = Mock()
        data_handle.get_data = Mock(return_value=data_array)
        data_handle.read_coefficients = Mock(return_value=actual_coefficients)
        adaptor.simulate(data_handle)

        actual = data_handle.set_results.call_args[0][1]
        expected = np.array([48])  # area zero
        assert np.allclose(actual, expected)

    def test_half_to_one_region_pass_through_time(self, regions_rect, regions_half_squares):
        """Convert from half a region to one region, pass through time
        """
        adaptor = RegionAdaptor('test-square-half')
        from_spec = Spec(
            name='test-var',
            dtype='float',
            dims=['rect', 'months'],
            coords={
                'rect': regions_rect,
                'months': list(range(12))
            }
        )
        adaptor.add_input(from_spec)
        to_spec = Spec(
            name='test-var',
            dtype='float',
            dims=['half_squares', 'months'],
            coords={
                'half_squares': regions_half_squares,
                'months': list(range(12))
            }
        )
        adaptor.add_output(to_spec)

        actual_coefficients = adaptor.generate_coefficients(from_spec, to_spec)
        expected = np.ones((1, 2)) / 2  # disaggregating coefficients
        np.testing.assert_allclose(actual_coefficients, expected, rtol=1e-3)

        data = np.ones((1, 12))  # area zero, months 1-12

        data_array = DataArray(from_spec, data)

        data_handle = Mock()
        data_handle.get_data = Mock(return_value=data_array)
        data_handle.read_coefficients = Mock(return_value=actual_coefficients)
        adaptor.simulate(data_handle)

        actual = data_handle.set_results.call_args[0][1]
        expected = np.ones((2, 12)) / 2  # areas a-b, months 1-12
        assert np.allclose(actual, expected)


def test_proportion(regions):
    """Sense-check proportion calculator
    """
    region_set = RegionSet('regions', regions)

    assert region_set.get_proportion(0, region_set[0]) == 1

    assert region_set.get_proportion(0, region_set[1]) == 0.5
    assert region_set.get_proportion(1, region_set[0]) == 1

    assert region_set.get_proportion(0, region_set[2]) == 1
    assert region_set.get_proportion(2, region_set[0]) == 0.5
    assert region_set.get_proportion(1, region_set[2]) == 1
    assert region_set.get_proportion(2, region_set[1]) == 0.25


class TestRegionSet():
    """Test creating, looking up, retrieving region sets
    """

    def test_serialise(self, regions):
        rset = RegionSet('test', regions)
        rset.description = 'my description'
        actual = rset.as_dict()
        expected = {'name': 'test',
                    'description': 'my description'}

        assert actual == expected

    def test_create(self, regions):
        rset = RegionSet('test', regions)
        assert rset.name == 'test'
        assert len(rset) == 3
        assert rset[0].name == 'unit'
        assert rset[1].name == 'half'
        assert rset[2].name == 'two'

    def test_get_names(self, regions):
        rset = RegionSet('test', regions)
        actual = rset.get_entry_names()
        expected = ['unit', 'half', 'two']
        assert actual == expected

    def test_as_features(self, regions_half_squares):
        """Retrieve regions as feature dicts
        """
        rset = RegionSet('test', regions_half_squares)
        actual = rset.as_features()
        expected = [
            {
                'type': 'Feature',
                'properties': {'name': 'a'},
                'geometry': {
                    'type': 'Polygon',
                    'coordinates': (
                        ((0.0, 0.0), (0.0, 1.0), (1.0, 1.0), (1.0, 0.0), (0.0, 0.0)),
                    )
                }
            },
            {
                'type': 'Feature',
                'properties': {'name': 'b'},
                'geometry': {
                    'type': 'Polygon',
                    'coordinates': (
                        ((0.0, 1.0), (0.0, 2.0), (1.0, 2.0), (1.0, 1.0), (0.0, 1.0)),
                    )
                }
            }
        ]
        assert actual == expected

    def test_centroids_as_features(self, regions_half_squares):
        """Retrieve centroids
        """
        actual = RegionSet('test', regions_half_squares).centroids_as_features()
        expected = [
            {
                'type': 'Feature',
                'properties': {'name': 'a'},
                'geometry': {
                    'type': 'Point',
                    'coordinates': (0.5, 0.5)
                }
            },
            {
                'type': 'Feature',
                'properties': {'name': 'b'},
                'geometry': {
                    'type': 'Point',
                    'coordinates': (0.5, 1.5)
                }
            },
        ]
        assert actual == expected

    def test_must_have_unique_names(self):
        with raises(AssertionError) as ex:
            RegionSet('test', [
                {
                    'name': 'a',
                    'feature': {
                        'type': 'Feature',
                        'properties': {'name': 'a'},
                        'geometry': {
                            'type': 'Polygon',
                            'coordinates': [[[0, 0], [0, 1], [1, 1], [1, 0]]]
                        }
                    }
                },
                {
                    'name': 'a',
                    'feature': {
                        'type': 'Feature',
                        'properties': {'name': 'a'},
                        'geometry': {
                            'type': 'Polygon',
                            'coordinates': [[[0, 1], [0, 2], [1, 2], [1, 1]]]
                        }
                    }
                }
            ])

        assert 'Region set must have uniquely named regions' in str(ex)


class TestRegionRegister():
    """Test creating registers, registering region sets, converting data
    """
    def test_create(self):
        register = NDimensionalRegister()
        assert register.names == []

        with raises(ValueError) as ex:
            register.get_entry('nonexistent')
        assert "ResolutionSet 'nonexistent' not registered" in str(ex)

    def test_convert_equal(self, register):
        data = np.ones(1)
        converted = register.convert(data, 'rect', 'rect_alt')
        np.testing.assert_equal(data, converted)

    def test_convert_to_half(self, register):
        data = np.ones(1)
        converted = register.convert(data, 'rect', 'half_squares')
        expected = np.array([0.5, 0.5])
        np.testing.assert_equal(converted, expected)

    def test_convert_from_half(self, register):
        data = np.ones(2) / 2
        converted = register.convert(data, 'half_squares', 'rect')
        expected = np.ones(1)
        np.testing.assert_equal(converted, expected)

        data = np.array([2, 3])
        converted = register.convert(data, 'half_squares', 'rect')
        expected = np.array([5])
        np.testing.assert_equal(converted, expected)

    def test_coverage_half(self, register):
        half_covered = register.get_entry('single_half_square')
        actual = half_covered.coverage
        expected = 1.0
        assert actual == expected

    def test_coverage_whole(self, register):
        rect = register.get_entry('rect')
        actual = rect.coverage
        expected = 2.0
        assert actual == expected

    def test_convert_to_half_not_covered(self, register, caplog):
        data = np.array([3])
        register.convert(data, 'rect', 'single_half_square')
        expected = "Coverage for 'rect' is 2 and does not match " \
                   "coverage for 'single_half_square' which is 1"

        assert expected in caplog.text

    def test_convert_from_half_not_covered(self, register, caplog):
        data = np.array([3])
        register.convert(data, 'single_half_square', 'rect')

        expected = "Coverage for 'single_half_square' is 1 and does not " \
                   "match coverage for 'rect' which is 2"

        assert expected in caplog.text

    def test_convert_square_to_triangle(self, register):
        data = np.array([1, 1])
        converted = register.convert(data, 'half_squares', 'half_triangles')
        expected = np.array([1, 1])
        np.testing.assert_equal(converted, expected)

        data = np.array([0, 1])
        converted = register.convert(data, 'half_squares', 'half_triangles')
        expected = np.array([0.25, 0.75])
        np.testing.assert_equal(converted, expected)

    def test_convert_triangle_to_square(self, register):
        data = np.array([1, 1])
        converted = register.convert(data, 'half_triangles', 'half_squares')
        expected = np.array([1, 1])
        np.testing.assert_equal(converted, expected)

        data = np.array([0.25, 0.75])
        converted = register.convert(data, 'half_triangles', 'half_squares')
        expected = np.array([0.375, 0.625])
        np.testing.assert_equal(converted, expected)


class TestGetCoefficients:

    def test_get_coefficients(self, register):
        actual = register.get_coefficients('half_triangles', 'half_squares')
        expected = np.array([[0.75, 0.25],
                             [0.25, 0.75]])
        np.testing.assert_equal(actual, expected)

    def test_coefs_should_map_to_themselves(self, register):
        actual = register.get_coefficients('half_triangles', 'half_triangles')
        expected = np.array([[1, 0],
                             [0, 1]])
        np.testing.assert_equal(actual, expected)



"""File-backed data store
"""
import glob
import os
from abc import abstractmethod
from logging import getLogger

import numpy as np  # type: ignore
import pandas  # type: ignore
import pyarrow as pa  # type: ignore
from smif.data_layer.abstract_data_store import DataStore
from smif.data_layer.data_array import DataArray
from smif.exception import SmifDataMismatchError, SmifDataNotFoundError


class FileDataStore(DataStore):
    """Abstract file data store
    """
    def __init__(self, base_folder):
        super().__init__()
        self.logger = getLogger(__name__)
        # extension for DataArray/list-of-dict data - override in implementations
        self.ext = ''
        # extension for bare numpy.ndarray data - override in implementations
        self.coef_ext = ''

        self.base_folder = str(base_folder)
        self.data_folder = str(os.path.join(self.base_folder, 'data'))
        self.data_folders = {}
        self.results_folder = str(os.path.join(self.base_folder, 'results'))
        data_folders = [
            'coefficients',
            'strategies',
            'initial_conditions',
            'interventions',
            'narratives',
            'scenarios',
            'strategies',
            'parameters'
        ]
        for folder in data_folders:
            dirname = os.path.join(self.data_folder, folder)
            # ensure each directory exists
            if not os.path.exists(dirname):
                msg = "Expected data folder at '{}' but it does does not exist"
                abs_path = os.path.abspath(dirname)
                raise SmifDataNotFoundError(msg.format(abs_path))
            self.data_folders[folder] = dirname

    # region Abstract methods
    @abstractmethod
    def _read_data_array(self, path, spec, timestep=None):
        """Read DataArray from file
        """

    @abstractmethod
    def _write_data_array(self, path, data_array, timestep=None):
        """Write DataArray to file
        """

    @abstractmethod
    def _read_list_of_dicts(self, path):
        """Read file to list[dict]
        """

    @abstractmethod
    def _write_list_of_dicts(self, path, data):
        """Write list[dict] to file
        """

    @abstractmethod
    def _read_ndarray(self, path):
        """Read numpy.ndarray
        """

    @abstractmethod
    def _write_ndarray(self, path, data, header=None):
        """Write numpy.ndarray
        """
    # endregion

    # region Data Array
    def read_scenario_variant_data(self, key, spec, timestep=None):
        path = os.path.join(self.data_folders['scenarios'], key)
        data = self._read_data_array(path, spec, timestep)
        data.validate_as_full()
        return data

    def write_scenario_variant_data(self, key, data, timestep=None):
        path = os.path.join(self.data_folders['scenarios'], key)
        self._write_data_array(path, data, timestep)

    def read_narrative_variant_data(self, key, spec, timestep=None):
        path = os.path.join(self.data_folders['narratives'], key)
        return self._read_data_array(path, spec, timestep)

    def write_narrative_variant_data(self, key, data, timestep=None):
        path = os.path.join(self.data_folders['narratives'], key)
        self._write_data_array(path, data, timestep)

    def read_model_parameter_default(self, key, spec):
        self.logger.debug("Trying to read model parameter default from key {}".format(key))
        path = os.path.join(self.data_folders['parameters'], key)
        data = self._read_data_array(path, spec)
        data.validate_as_full()
        return data

    def write_model_parameter_default(self, key, data):
        path = os.path.join(self.data_folders['parameters'], key)
        self._write_data_array(path, data)
    # endregion

    # region Interventions
    def read_interventions(self, keys):
        all_interventions = []
        for key in keys:
            path = os.path.join(self.data_folders['interventions'], key)
            interventions = self._read_list_of_dicts(path)
            all_interventions.extend(interventions)

        seen = set()
        dups = set()

        for intervention in all_interventions:
            try:
                name = intervention['name']
            except KeyError:
                msg = "Could not find `name` key in {} for {}"
                raise KeyError(msg.format(intervention, keys))
            if name in seen:
                dups.add(name)
            else:
                seen.add(name)

        if dups:
            name = dups.pop()
            msg = "An entry for intervention {} already exists. Also found duplicates for {}"
            raise ValueError(msg.format(name, dups))

        return {
            intervention['name']: _nest_keys(intervention)
            for intervention in all_interventions
        }

    def write_interventions(self, key, interventions):
        # convert dict[str, dict] to list[dict]
        data = [
            _unnest_keys(intervention)
            for intervention in interventions.values()
        ]
        path = os.path.join(self.data_folders['interventions'], key)
        self._write_list_of_dicts(path, data)

    def read_strategy_interventions(self, strategy):
        path = os.path.join(self.data_folders['strategies'], strategy['filename'])
        return self._read_list_of_dicts(path)

    def read_initial_conditions(self, keys):
        conditions = []
        for key in keys:
            path = os.path.join(self.data_folder, 'initial_conditions', key)
            data = self._read_list_of_dicts(path)
            conditions.extend(data)
        return conditions

    def write_initial_conditions(self, key, initial_conditions):
        path = os.path.join(self.data_folders['initial_conditions'], key)
        self._write_list_of_dicts(path, initial_conditions)
    # endregion

    # region State
    def read_state(self, modelrun_name, timestep, decision_iteration=None):
        path = self._get_state_path(modelrun_name, timestep, decision_iteration)
        try:
            state = self._read_list_of_dicts(path)
        except FileNotFoundError:
            msg = "State file does not exist for timestep {} and iteration {}"
            raise SmifDataNotFoundError(msg.format(timestep, decision_iteration))
        return state

    def write_state(self, state, modelrun_name, timestep=None, decision_iteration=None):
        path = self._get_state_path(modelrun_name, timestep, decision_iteration)
        os.makedirs(os.path.dirname(path), exist_ok=True)
        self._write_list_of_dicts(path, state)

    def _get_state_path(self, modelrun_name, timestep=None, decision_iteration=None):
        """Compose a unique filename for state file:
                state_{timestep|0000}[_decision_{iteration}].{ext}
        """
        if timestep is None:
            timestep = '0000'

        if decision_iteration is None:
            separator = ''
            decision_iteration = ''
        else:
            separator = '_decision_'

        filename = 'state_{}{}{}.{}'.format(timestep, separator, decision_iteration, self.ext)
        path = os.path.join(self.results_folder, modelrun_name, filename)

        return path
    # endregion

    # region Conversion coefficients
    def read_coefficients(self, source_dim, destination_dim):
        results_path = self._get_coefficients_path(source_dim, destination_dim)
        try:
            return self._read_ndarray(results_path)
        except FileNotFoundError:
            msg = "Could not find the coefficients file for %s to %s"
            self.logger.warning(msg, source_dim, destination_dim)
            raise SmifDataNotFoundError(msg.format(source_dim, destination_dim))

    def write_coefficients(self, source_dim, destination_dim, data):
        results_path = self._get_coefficients_path(source_dim, destination_dim)
        header = "Conversion coefficients {}:{}".format(source_dim, destination_dim)
        self._write_ndarray(results_path, data, header)

    def _get_coefficients_path(self, source_dim, destination_dim):
        path = os.path.join(
            self.data_folders['coefficients'],
            "{}.{}.{}".format(
                source_dim,
                destination_dim,
                self.coef_ext
            )
        )
        return path
    # endregion

    # region Results

    def read_results(self, modelrun_id, model_name, output_spec, timestep,
                     decision_iteration=None):
        if timestep is None:
            raise ValueError("You must pass a timestep argument")

        results_path = self._get_results_path(
            modelrun_id, model_name, output_spec.name,
            timestep, decision_iteration
        )

        try:
            return self._read_data_array(results_path, output_spec)
        except FileNotFoundError:
            key = str([modelrun_id, model_name, output_spec.name, timestep,
                       decision_iteration])
            raise SmifDataNotFoundError("Could not find results for {}".format(key))

    def write_results(self, data_array, modelrun_id, model_name, timestep=None,
                      decision_iteration=None):
        if timestep is None:
            raise NotImplementedError()

        if timestep:
            assert isinstance(timestep, int), "Timestep must be an integer"
        if decision_iteration:
            assert isinstance(decision_iteration, int), "Decision iteration must be an integer"

        results_path = self._get_results_path(
            modelrun_id, model_name, data_array.name,
            timestep, decision_iteration
        )
        os.makedirs(os.path.dirname(results_path), exist_ok=True)
        self._write_data_array(results_path, data_array)

    def available_results(self, modelrun_name):
        """List available results for a given model run

        See _get_results_path for path construction.

        On the pattern of:
            results/<modelrun_name>/<model_name>/
            decision_<id>/
            output_<output_name>_timestep_<timestep>.csv
        """
        paths = glob.glob(os.path.join(
            self.results_folder, modelrun_name, "*", "*", "*.{}".format(self.ext)))
        # (timestep, decision_iteration, model_name, output_name)
        results_keys = []
        for path in paths:
            timestep, decision_iteration, model_name, output_name = \
                self._parse_results_path(path)
            results_keys.append(
                (timestep, decision_iteration, model_name, output_name)
            )
        return results_keys

    def _get_results_path(self, modelrun_id, model_name, output_name, timestep,
                          decision_iteration=None):
        """Return path to filename for a given output without file extension

        On the pattern of:
            results/<modelrun_name>/<model_name>/
            decision_<id>/
            output_<output_name>_timestep_<timestep>.csv

        Parameters
        ----------
        modelrun_id : str
        model_name : str
        output_name : str
        timestep : str or int
        decision_iteration : int, optional

        Returns
        -------
        path : strs
        """
        if decision_iteration is None:
            decision_iteration = 'none'

        path = os.path.join(
            self.results_folder, modelrun_id, model_name,
            "decision_{}".format(decision_iteration),
            "output_{}_timestep_{}.{}".format(output_name, timestep, self.ext)
        )
        return path

    def _parse_results_path(self, path):
        """Return result metadata for a given result path

        On the pattern of:
            results/<modelrun_name>/<model_name>/
            decision_<id>/
            output_<output_name>_timestep_<timestep>.<ext>

        Parameters
        ----------
        path : str

        Returns
        -------
        tuple : (timestep, decision_iteration, model_name, output_name)
        """
        # split to last directories and filename
        model_name, decision_str, output_str = path.split(os.sep)[-3:]
        # trim "decision_"
        decision_iteration = int(decision_str[9:])
        # trim "output_" [...]
        output_str_trimmed = output_str[7:]
        # trim extension
        output_str_trimmed = output_str_trimmed.replace(".{}".format(self.ext), "")
        # pick (str) output and (integer) timestep
        output_name, timestep_str = output_str_trimmed.split("_timestep_")
        timestep = int(timestep_str)

        return (timestep, decision_iteration, model_name, output_name)
    # endregion

    def _filter_on_timestep(self, timestep, dataframe, path, spec):
        if timestep is not None:
            if 'timestep' not in dataframe.columns:
                if 'timestep' not in dataframe.index.names:
                    msg = "Data for '{name}' expected a column called 'timestep', instead " + \
                          "got data columns {data_columns} and index names {index_names} " + \
                          "while reading from {path}"
                    raise SmifDataMismatchError(msg.format(
                        data_columns=dataframe.columns.values.tolist(),
                        index_names=dataframe.index.names,
                        name=spec.name,
                        path=path))
                dataframe = dataframe.reset_index(level='timestep')

            dataframe = dataframe[dataframe.timestep == timestep]

            if dataframe.empty:
                raise SmifDataNotFoundError(
                    "Data for '{}' not found for timestep {}".format(spec.name, timestep))

            dataframe = dataframe.drop('timestep', axis=1)

        return dataframe


class CSVDataStore(FileDataStore):
    """CSV text file data store
    """
    def __init__(self, base_folder):
        super().__init__(base_folder)
        self.ext = 'csv'
        self.coef_ext = 'txt.gz'

    def _read_data_array(self, path, spec, timestep=None):
        """Read DataArray from file
        """
        try:
            dataframe = pandas.read_csv(path)
        except FileNotFoundError:
            raise SmifDataNotFoundError

        dataframe = self._filter_on_timestep(timestep, dataframe, path, spec)

        if spec.dims:
            data_array = DataArray.from_df(spec, dataframe)
        else:
            # zero-dimensional case (scalar)
            data = dataframe[spec.name]
            if data.shape != (1,):
                msg = "Data for '{}' should contain a single value, instead got {} while " + \
                      "reading from {}"
                raise SmifDataMismatchError(msg.format(spec.name, len(data), path))
            data_array = DataArray(spec, data.iloc[0])
        return data_array

    def _write_data_array(self, path, data_array, timestep=None):
        """Write DataArray to file
        """
        dataframe = data_array.as_df()
        if timestep is not None:
            dataframe['timestep'] = timestep
        dataframe.reset_index().to_csv(path, index=False)

    def _read_list_of_dicts(self, path):
        """Read file to list[dict]
        """
        return pandas.read_csv(path).to_dict('records')

    def _write_list_of_dicts(self, path, data):
        """Write list[dict] to file
        """
        pandas.DataFrame.from_records(data).to_csv(path, index=False)

    def _read_ndarray(self, path):
        """Read numpy.ndarray
        """
        try:
            return np.loadtxt(path)
        except OSError:
            raise FileNotFoundError(path)

    def _write_ndarray(self, path, data, header=None):
        """Write numpy.ndarray
        """
        np.savetxt(path, data, header=header)


class ParquetDataStore(FileDataStore):
    """Binary file data store
    """
    def __init__(self, base_folder):
        super().__init__(base_folder)
        self.ext = 'parquet'
        self.coef_ext = 'npy'

    def _read_parquet_data_array(self, path, spec, timestep=None):

        dataframe = pandas.read_parquet(path, engine='pyarrow')
        dataframe = self._filter_on_timestep(timestep, dataframe, path, spec)

        if spec.dims:
            data_array = DataArray.from_df(spec, dataframe)
        else:
            # zero-dimensional case (scalar)
            data = dataframe[spec.name]
            if data.shape != (1,):
                msg = "Expected single value, found {} in {}"
                raise SmifDataMismatchError(msg.format(list(data.shape), path))
            data_array = DataArray(spec, data.iloc[0])

        return data_array

    def _read_data_array(self, path, spec, timestep=None):
        """Read DataArray from file
        """
        try:
            data_array = self._read_parquet_data_array(path, spec, timestep)
        except (pa.lib.ArrowIOError, OSError) as ex:
            msg = "Could not find data for {} at {}"
            raise SmifDataNotFoundError(msg.format(spec.name, path)) from ex
        return data_array

    def _write_data_array(self, path, data_array, timestep=None):
        """Write DataArray to file
        """
        dataframe = data_array.as_df()
        if timestep is not None:
            dataframe['timestep'] = timestep
        dataframe.to_parquet(path, engine='pyarrow', compression='gzip')

    def _read_list_of_dicts(self, path):
        """Read file to list[dict]
        """
        try:
            return pandas.read_parquet(path, engine='pyarrow').to_dict('records')
        except pa.lib.ArrowIOError as ex:
            msg = "Unable to read file at {}"
            raise SmifDataNotFoundError(msg.format(path)) from ex

    def _write_list_of_dicts(self, path, data):
        """Write list[dict] to file
        """
        pandas.DataFrame.from_records(data).to_parquet(path, engine='pyarrow')

    def _read_ndarray(self, path):
        """Read numpy.ndarray
        """
        try:
            return np.load(path)
        except OSError:
            raise FileNotFoundError(path)

    def _write_ndarray(self, path, data, header=None):
        """Write numpy.ndarray
        """
        np.save(path, data)


def _nest_keys(intervention):
    nested = {}
    for key, value in intervention.items():
        if key.endswith(('_value', '_unit')):
            new_key, sub_key = key.rsplit(sep="_", maxsplit=1)
            if new_key in nested:
                if not isinstance(nested[new_key], dict):
                    msg = "Duplicate heading in csv data: {}"
                    raise ValueError(msg.format(new_key))
                else:
                    nested[new_key].update({sub_key: value})
            else:
                nested[new_key] = {sub_key: value}
        else:
            if key in nested:
                msg = "Duplicate heading in csv data: {}"
                raise ValueError(msg.format(new_key))
            else:
                nested[key] = value
    return nested


def _unnest_keys(intervention):
    unnested = {}
    for key, value in intervention.items():
        if isinstance(value, dict):
            for sub_key, sub_value in value:
                unnested["{}_{}".format(key, sub_key)] = sub_value
        else:
            unnested[key] = value
    return unnested



"""Handles conversion between the sets of regions used in the `SosModel`
"""
from collections import namedtuple

from rtree import index  # type: ignore
from shapely.geometry import mapping, shape  # type: ignore
from shapely.validation import explain_validity  # type: ignore
from smif.convert.adaptor import Adaptor
from smif.convert.register import NDimensionalRegister, ResolutionSet

__author__ = "Will Usher, Tom Russell"
__copyright__ = "Will Usher, Tom Russell"
__license__ = "mit"


class RegionAdaptor(Adaptor):
    """Convert regions, assuming uniform distributions where necessary
    """
    def generate_coefficients(self, from_spec, to_spec):
        """Generate conversion coefficients for spatial dimensions

        Assumes that the Coordinates elements contain a 'feature' key whose value corresponds
        to a GDAL vector feature represented as a dict, for example as returned by a `fiona`
        reader.
        """
        # find dimensions to convert
        from_dim, to_dim = self.get_convert_dims(from_spec, to_spec)
        # get dimension coordinates
        from_coords = from_spec.dim_coords(from_dim)
        to_coords = to_spec.dim_coords(to_dim)
        # create RegionSets from Coordinates
        from_set = RegionSet(from_dim, from_coords.elements)
        to_set = RegionSet(to_dim, to_coords.elements)
        # register RegionSets
        register = NDimensionalRegister()
        register.register(from_set)
        register.register(to_set)
        # use NDimensionalRegister to get coefficients
        coefficients = register.get_coefficients(from_dim, to_dim)
        return coefficients


NamedShape = namedtuple('NamedShape', ['name', 'shape'])


class RegionSet(ResolutionSet):
    """Hold a set of regions, spatially indexed for ease of lookup when
    constructing conversion matrices.

    Parameters
    ----------
    set_name : str
        Name to use as identifier for this set of regions
    elements: iterable
        Iterable (probably a list or a reader handle)
        of fiona feature records e.g. the 'features' entry of
        a GeoJSON collection

    """
    def __init__(self, set_name, elements):
        super().__init__()
        self.name = set_name
        self._regions = []
        self.data = [e['feature'] for e in elements]

        self._idx = index.Index()
        for pos, region in enumerate(self._regions):
            self._idx.insert(pos, region.shape.bounds)

    @property
    def data(self):
        return self._regions

    @data.setter
    def data(self, value):
        names = {}
        for region in value:
            name = region['properties']['name']
            if name in names:
                msg = "Region set must have uniquely named regions - {} duplicated"
                raise AssertionError(msg.format(name))
            names[name] = True
            self._regions.append(
                NamedShape(
                    name,
                    shape(region['geometry'])
                )
            )

    def get_entry_names(self):
        return [region.name for region in self.data]

    def as_features(self):
        """Get the regions as a list of feature dictionaries

        Returns
        -------
        list
            A list of GeoJSON-style dicts
        """
        return [
            {
                'type': 'Feature',
                'geometry': mapping(region.shape),
                'properties': {
                    'name': region.name
                }
            }
            for region in self._regions
        ]

    def centroids_as_features(self):
        """Get the region centroids as a list of feature dictionaries

        Returns
        -------
        list
            A list of GeoJSON-style dicts, with Point features corresponding to
            region centroids
        """
        return [
            {
                'type': 'Feature',
                'geometry': mapping(region.shape.centroid),
                'properties': {
                    'name': region.name
                }
            }
            for region in self._regions
        ]

    def intersection(self, to_entry):
        """Return the set of regions intersecting with the bounds of `to_entry`
        """
        bounds = to_entry.shape.bounds
        return [x for x in self._idx.intersection(bounds)]

    def get_proportion(self, from_idx, entry_b):
        """Calculate the proportion of shape a that intersects with shape b
        """
        entry_a = self.data[from_idx]
        if self.check_valid_shape(entry_a.shape):
            if self.check_valid_shape(entry_b.shape):
                intersection = entry_a.shape.intersection(entry_b.shape)
                return intersection.area / entry_a.shape.area
            else:
                raise RuntimeError("Shape {} is not valid".format(entry_b.name))
        else:
            raise RuntimeError("Shape {} from {} is not valid".format(entry_a.name, self.name))

    def check_valid_shape(self, shape):
        if not shape.is_valid:
            validity = explain_validity(shape)
            print("Shape is not valid. Explanation: %s", validity)
            return False
        else:
            return True

    @staticmethod
    def get_bounds(entry):
        return entry.shape.bounds

    @property
    def coverage(self):
        return sum([x.shape.area for x in self.data])

    def __getitem__(self, key):
        return self._regions[key]

    def __len__(self):
        return len(self._regions)



"""Test dependency representation
"""
from copy import copy
from unittest.mock import Mock

from pytest import fixture, raises
from smif.metadata import RelativeTimestep, Spec
from smif.model.dependency import Dependency


@fixture(scope='function')
def dep():
    """Dependency with mocked models
    """
    source_model = Mock()
    source_model.name = 'source_model'
    source_spec = Spec(
        name='source',
        dtype='float',
        dims=['x', 'y'],
        coords={'x': [0, 1], 'y': [0, 1]},
    )
    sink_model = Mock()
    sink_model.name = 'sink_model'
    sink_spec = Spec(
        name='sink',
        dtype='float',
        dims=['x', 'y'],
        coords={'x': [0, 1], 'y': [0, 1]},
    )
    return Dependency(source_model, source_spec, sink_model, sink_spec)


def test_create():
    """Create with models and specs, access properties
    """
    source_model = Mock()
    source_spec = Spec(
        name='source',
        dtype='float',
        dims=['x', 'y'],
        coords={'x': [0, 1], 'y': [0, 1]}
    )
    sink_model = Mock()
    sink_spec = Spec(
        name='sink',
        dtype='float',
        dims=['x', 'y'],
        coords={'x': [0, 1], 'y': [0, 1]},
    )
    dep = Dependency(source_model, source_spec, sink_model, sink_spec,
                     RelativeTimestep.PREVIOUS)
    assert dep.source_model == source_model
    assert dep.source == source_spec
    assert dep.sink_model == sink_model
    assert dep.sink == sink_spec
    assert dep.timestep == RelativeTimestep.PREVIOUS

    # default to current timestep
    dep = Dependency(source_model, source_spec, sink_model, sink_spec)
    assert dep.timestep == RelativeTimestep.CURRENT


def test_create_with_identical_meta():
    source = Spec(
        name='source',
        dtype='float',
        dims=['x', 'y'],
        coords={'x': [0, 1], 'y': [0, 1]},
        unit='m'
    )
    a_sink = Spec(
        name='sink',  # differs but that's okay
        dtype='float',
        dims=['x', 'y'],
        coords={'x': [0, 1], 'y': [0, 1]},
        unit='m'
    )
    Dependency(Mock(), source, Mock(), a_sink)

    b_sink = Spec(
        name='sink',
        dtype='int',  # differs
        dims=['x', 'y'],
        coords={'x': [0, 1], 'y': [0, 1]},
        unit='m'
    )
    with raises(ValueError) as ex:
        Dependency(Mock(), source, Mock(), b_sink)
    assert 'mismatched dtype' in str(ex)

    c_sink = Spec(
        name='sink',
        dtype='float',
        dims=['x', 'z'],  # differs
        coords={'x': [0, 1], 'z': [0, 1]},
        unit='m'
    )
    with raises(ValueError) as ex:
        Dependency(Mock(), source, Mock(), c_sink)
    assert 'mismatched dims' in str(ex)

    d_sink = Spec(
        name='sink',
        dtype='float',
        dims=['x', 'y'],
        coords={'x': [0, 1], 'y': [1, 2]},  # differs
        unit='m'
    )
    with raises(ValueError) as ex:
        Dependency(Mock(), source, Mock(), d_sink)
    assert 'mismatched coords' in str(ex)

    e_sink = Spec(
        name='sink',
        dtype='float',
        dims=['x', 'y'],
        coords={'x': [0, 1], 'y': [0, 1]},
        unit='km'  # differs
    )
    with raises(ValueError) as ex:
        Dependency(Mock(), source, Mock(), e_sink)
    assert 'mismatched unit' in str(ex)


def test_repr(dep):
    actual = repr(dep)
    expected = "<Dependency({}, {}, {}, {}, {})>".format(
        dep.source_model, dep.source, dep.sink_model, dep.sink, dep.timestep)
    assert actual == expected


def test_as_dict(dep):
    assert dep.as_dict() == {
        'source': 'source_model',
        'source_output': 'source',
        'sink_input': 'sink',
        'sink': 'sink_model',
        'timestep': 'CURRENT'
    }


def test_equality(dep):
    a = dep
    b = copy(dep)
    assert a == b

    c = copy(dep)
    c.source_model = Mock()
    assert a != c

    another_spec = Spec(
        name='another',
        dtype='float',
        dims=['z'],
        coords={'z': [0, 1]}
    )
    d = copy(dep)
    d.source = another_spec
    assert a != d

    e = copy(dep)
    e.sink = another_spec
    assert a != e

    f = copy(dep)
    f.sink_model = Mock()
    assert a != f

    g = copy(dep)
    g.timestep = RelativeTimestep.BASE
    assert a != g



from unittest.mock import Mock

from pytest import fixture
from smif.metadata import Spec
from smif.model.model import ScenarioModel


@fixture(scope='function')
def scenario_model():
    """Sample scenario model
    """
    scenario_model = ScenarioModel('population')
    scenario_model.scenario = 'High Population (ONS)'
    scenario_model.description = 'The High ONS Forecast for UK population out to 2050'
    scenario_model.add_output(
        Spec(
            name='population_count',
            dims=['LSOA'],
            coords={'LSOA': ['a', 'b', 'c']},
            dtype='int',
            unit='people'
        )
    )
    return scenario_model


@fixture(scope='function')
def scenario_model_dict():
    """Dict serialisation of sample scenario model
    """
    return {
        'name': 'population',
        'scenario': 'High Population (ONS)',
        'description': 'The High ONS Forecast for UK population out to 2050',
        'outputs': [
            {
                'name': 'population_count',
                'dims': ['LSOA'],
                'coords': {'LSOA': ['a', 'b', 'c']},
                'dtype': 'int',
                'unit': 'people',
                'abs_range': None,
                'exp_range': None,
                'description': None
            }
        ]
    }


class TestScenarioModel(object):
    """ScenarioModel should represent data available from pre-computed datasets
    """
    def test_construct(self, scenario_model):
        assert scenario_model.name == 'population'
        assert scenario_model.scenario == 'High Population (ONS)'
        assert scenario_model.description == \
            'The High ONS Forecast for UK population out to 2050'
        assert scenario_model.outputs == {
            'population_count': Spec.from_dict({
                'name': 'population_count',
                'dims': ['LSOA'],
                'coords': {'LSOA': ['a', 'b', 'c']},
                'dtype': 'int',
                'unit': 'people'
            })
        }

    def test_from_dict(self, scenario_model, scenario_model_dict):
        actual = ScenarioModel.from_dict(scenario_model_dict)
        expected = scenario_model
        assert actual.name == expected.name
        assert actual.scenario == expected.scenario
        assert actual.description == expected.description
        assert actual.outputs == expected.outputs

    def test_from_dict_option_description(self, scenario_model_dict):
        del scenario_model_dict['description']
        scenario_model = ScenarioModel.from_dict(scenario_model_dict)
        assert scenario_model.description == ''

    def test_as_dict(self, scenario_model, scenario_model_dict):
        """Serialise metadata to dict
        """
        actual = scenario_model.as_dict()
        expected = scenario_model_dict
        assert actual == expected

    def test_as_dict_multi_output(self, scenario_model, scenario_model_dict):
        # with additional output
        scenario_model.add_output(
            Spec.from_dict({
                'name': 'population_density',
                'dims': ['LSOA'],
                'coords': {'LSOA': ['a', 'b', 'c']},
                'dtype': 'int',
                'unit': 'people / kilometer ** 2'
            })
        )
        scenario_model_dict['outputs'].append({
            'name': 'population_density',
            'dims': ['LSOA'],
            'coords': {'LSOA': ['a', 'b', 'c']},
            'dtype': 'int',
            'unit': 'people / kilometer ** 2',
            'abs_range': None,
            'exp_range': None,
            'description': None
        })

        actual = scenario_model.as_dict()
        expected = scenario_model_dict

        # equivalent up to order of outputs
        actual['outputs'].sort(key=lambda p: p['name'])
        expected['outputs'].sort(key=lambda p: p['name'])

        assert actual == expected

    def test_scenario_data(self, scenario_model):
        """Scenario model simulate method should be a no-op
        """
        data_handle = Mock()
        actual = scenario_model.simulate(data_handle)
        assert actual is data_handle
        data_handle.assert_not_called()



"""File-backed config store
"""
import copy
import os
from logging import getLogger

from ruamel.yaml import YAML  # type: ignore
from smif.data_layer.abstract_config_store import ConfigStore
from smif.data_layer.validate import (validate_sos_model_config,
                                      validate_sos_model_format)
from smif.exception import (SmifDataExistsError, SmifDataMismatchError,
                            SmifDataNotFoundError)


class YamlConfigStore(ConfigStore):
    """Config backend saving to YAML configuration files.

    Arguments
    ---------
    base_folder: str
        The path to the configuration and data files
    """
    def __init__(self, base_folder, validation=False):
        super().__init__()
        self.logger = getLogger(__name__)
        self.validation = validation

        self.base_folder = str(base_folder)
        self.config_folder = str(os.path.join(self.base_folder, 'config'))
        self.config_folders = {}
        config_folders = [
            'dimensions',
            'model_runs',
            'scenarios',
            'sector_models',
            'sos_models',
        ]
        for folder in config_folders:
            dirname = os.path.join(self.config_folder, folder)
            # ensure each directory exists
            if not os.path.exists(dirname):
                msg = "Expected configuration folder at '{}' but it does not exist"
                abs_path = os.path.abspath(dirname)
                raise SmifDataNotFoundError(msg.format(abs_path))

            self.config_folders[folder] = dirname

        # cache results of reading project_config (invalidate on write)
        self._project_config_cache_invalid = True
        # MUST ONLY access through self.read_project_config()
        self._project_config_cache = None

        # ensure project config file exists
        try:
            self.read_project_config()
        except FileNotFoundError:
            # write empty config if none found
            self._write_project_config({})

    def read_project_config(self):
        """Read the project configuration

        Returns
        -------
        dict
            The project configuration
        """
        if self._project_config_cache_invalid:

            self._project_config_cache = _read_yaml_file(self.base_folder, 'project')
            self._project_config_cache_invalid = False
        return copy.deepcopy(self._project_config_cache)

    def _write_project_config(self, data):
        """Write the project configuration

        Argument
        --------
        data: dict
            The project configuration
        """
        self._project_config_cache_invalid = True
        self._project_config_cache = None
        _write_yaml_file(self.base_folder, 'project', data)

    def _read_config(self, config_type, config_name):
        """Read config item - used by decorators for existence/consistency checks
        """
        if config_type == 'scenario':
            return self.read_scenario(config_name)
        else:
            raise NotImplementedError(
                "Cannot read %s:%s through generic method." % (config_type, config_name))

    # region Model runs
    def read_model_runs(self):
        names = _read_filenames_in_dir(self.config_folders['model_runs'], '.yml')
        model_runs = [self.read_model_run(name) for name in names]
        return model_runs

    def read_model_run(self, model_run_name):
        _assert_file_exists(self.config_folders, 'model_run', model_run_name)
        modelrun_config = self._read_model_run(model_run_name)
        del modelrun_config['strategies']
        return modelrun_config

    def _read_model_run(self, model_run_name):
        return _read_yaml_file(self.config_folders['model_runs'], model_run_name)

    def _overwrite_model_run(self, model_run_name, model_run):
        _write_yaml_file(self.config_folders['model_runs'], model_run_name, model_run)

    def write_model_run(self, model_run):
        _assert_file_not_exists(self.config_folders, 'model_run', model_run['name'])
        config = copy.copy(model_run)
        config['strategies'] = []
        _write_yaml_file(self.config_folders['model_runs'], config['name'], config)

    def update_model_run(self, model_run_name, model_run):
        if model_run['name'] != model_run_name:
            raise SmifDataMismatchError(
                "Model run name '%s' must match '%s'" % (model_run_name, model_run['name']))
        _assert_file_exists(self.config_folders, 'model_run', model_run_name)
        prev = self._read_model_run(model_run_name)
        config = copy.copy(model_run)
        config['strategies'] = prev['strategies']
        self._overwrite_model_run(model_run_name, config)

    def delete_model_run(self, model_run_name):
        _assert_file_exists(self.config_folders, 'model_run', model_run_name)
        os.remove(os.path.join(self.config_folders['model_runs'], model_run_name + '.yml'))
    # endregion

    # region System-of-system models
    def read_sos_models(self):
        names = _read_filenames_in_dir(self.config_folders['sos_models'], '.yml')
        sos_models = [self.read_sos_model(name) for name in names]
        return sos_models

    def read_sos_model(self, sos_model_name):
        _assert_file_exists(self.config_folders, 'sos_model', sos_model_name)

        data = _read_yaml_file(self.config_folders['sos_models'], sos_model_name)
        if self.validation:
            validate_sos_model_format(data)
        return data

    def write_sos_model(self, sos_model):
        _assert_file_not_exists(self.config_folders, 'sos_model', sos_model['name'])
        _write_yaml_file(self.config_folders['sos_models'], sos_model['name'], sos_model)

    def update_sos_model(self, sos_model_name, sos_model):
        if sos_model['name'] != sos_model_name:
            raise SmifDataMismatchError(
                "SoSModel name '%s' must match '%s'" % (sos_model_name, sos_model['name']))
        _assert_file_exists(self.config_folders, 'sos_model', sos_model_name)
        if self.validation:
            validate_sos_model_config(
                sos_model,
                self.read_models(),
                self.read_scenarios(),
            )
        _write_yaml_file(self.config_folders['sos_models'], sos_model['name'], sos_model)

    def delete_sos_model(self, sos_model_name):
        _assert_file_exists(self.config_folders, 'sos_model', sos_model_name)
        os.remove(os.path.join(self.config_folders['sos_models'], sos_model_name + '.yml'))
    # endregion

    # region Models
    def read_models(self):
        names = _read_filenames_in_dir(self.config_folders['sector_models'], '.yml')
        models = [self.read_model(name) for name in names]
        return models

    def read_model(self, model_name):
        _assert_file_exists(self.config_folders, 'sector_model', model_name)

        model = _read_yaml_file(self.config_folders['sector_models'], model_name)
        return model

    def write_model(self, model):
        _assert_file_not_exists(self.config_folders, 'sector_model', model['name'])
        model = copy.deepcopy(model)
        if model['interventions']:
            self.logger.warning("Ignoring interventions")
            model['interventions'] = []

        model = _skip_coords(model, ('inputs', 'outputs', 'parameters'))
        _write_yaml_file(self.config_folders['sector_models'], model['name'], model)

    def update_model(self, model_name, model):
        if model['name'] != model_name:
            raise SmifDataMismatchError(
                "Model name '%s' must match '%s'" % (model_name, model['name']))
        _assert_file_exists(self.config_folders, 'sector_model', model_name)
        model = copy.deepcopy(model)

        # ignore interventions and initial conditions which the app doesn't handle
        if model['interventions'] or model['initial_conditions']:

            old_model = _read_yaml_file(self.config_folders['sector_models'], model['name'])

        if model['interventions']:
            self.logger.warning("Ignoring interventions write")
            model['interventions'] = old_model['interventions']

        if model['initial_conditions']:
            self.logger.warning("Ignoring initial conditions write")
            model['initial_conditions'] = old_model['initial_conditions']

        model = _skip_coords(model, ('inputs', 'outputs', 'parameters'))

        _write_yaml_file(self.config_folders['sector_models'], model['name'], model)

    def delete_model(self, model_name):
        _assert_file_exists(self.config_folders, 'sector_model', model_name)
        os.remove(
            os.path.join(self.config_folders['sector_models'], model_name + '.yml'))
    # endregion

    # region Scenarios
    def read_scenarios(self):
        scenario_names = _read_filenames_in_dir(self.config_folders['scenarios'], '.yml')
        return [self.read_scenario(name) for name in scenario_names]

    def read_scenario(self, scenario_name):
        _assert_file_exists(self.config_folders, 'scenario', scenario_name)

        scenario = _read_yaml_file(self.config_folders['scenarios'], scenario_name)
        return scenario

    def write_scenario(self, scenario):
        _assert_file_not_exists(self.config_folders, 'scenario', scenario['name'])
        scenario = _skip_coords(scenario, ['provides'])
        _write_yaml_file(self.config_folders['scenarios'], scenario['name'], scenario)

    def update_scenario(self, scenario_name, scenario):
        _assert_file_exists(self.config_folders, 'scenario', scenario_name)
        scenario = _skip_coords(scenario, ['provides'])
        _write_yaml_file(self.config_folders['scenarios'], scenario['name'], scenario)

    def delete_scenario(self, scenario_name):
        _assert_file_exists(self.config_folders, 'scenario', scenario_name)
        os.remove(
            os.path.join(self.config_folders['scenarios'], "{}.yml".format(scenario_name)))
    # endregion

    # region Scenario Variants
    def read_scenario_variants(self, scenario_name):
        scenario = self.read_scenario(scenario_name)
        return scenario['variants']

    def read_scenario_variant(self, scenario_name, variant_name):
        variants = self.read_scenario_variants(scenario_name)
        return _pick_from_list(variants, variant_name)

    def write_scenario_variant(self, scenario_name, variant):
        scenario = self.read_scenario(scenario_name)
        scenario['variants'].append(variant)
        self.update_scenario(scenario_name, scenario)

    def update_scenario_variant(self, scenario_name, variant_name, variant):
        scenario = self.read_scenario(scenario_name)
        v_idx = _idx_in_list(scenario['variants'], variant_name)
        scenario['variants'][v_idx] = variant
        self.update_scenario(scenario_name, scenario)

    def delete_scenario_variant(self, scenario_name, variant_name):
        scenario = self.read_scenario(scenario_name)
        v_idx = _idx_in_list(scenario['variants'], variant_name)
        del scenario['variants'][v_idx]
        self.update_scenario(scenario_name, scenario)
    # endregion

    # region Narratives
    def read_narrative(self, sos_model_name, narrative_name):
        sos_model = self.read_sos_model(sos_model_name)
        narrative = _pick_from_list(sos_model['narratives'], narrative_name)
        if not narrative:
            msg = "Narrative '{}' not found in '{}'"
            raise SmifDataNotFoundError(msg.format(narrative_name, sos_model_name))
        return narrative
    # endregion

    # region Strategies
    def read_strategies(self, modelrun_name):
        model_run_config = self._read_model_run(modelrun_name)
        return model_run_config['strategies']

    def write_strategies(self, modelrun_name, strategies):
        model_run = self._read_model_run(modelrun_name)
        model_run['strategies'] = strategies
        self._overwrite_model_run(modelrun_name, model_run)
    # endregion


def _read_yaml_file(directory, name):
    """Read yaml config file into plain data (lists, dicts and simple values)

    Parameters
    ----------
    directory : str
    name : str
    """
    path = os.path.join(directory, "{}.yml".format(name))
    with open(path, 'r') as file_handle:
        return YAML().load(file_handle)


def _write_yaml_file(directory, name, data):
    """Write plain data to a file as yaml

    Arguments
    ---------
    directory: str
        Path to directory
    name: str
        Name of config item (filename without .yml extension)
    data
        Data to be written to the file
    """
    path = os.path.join(directory, "{}.yml".format(name))
    with open(path, 'w') as file_handle:
        yaml = YAML()
        yaml.default_flow_style = False
        yaml.allow_unicode = True
        return yaml.dump(data, file_handle)


def _assert_file_exists(file_dir, dtype, name):
    if not _file_exists(file_dir, dtype, name):
        raise SmifDataNotFoundError("%s '%s' not found" % (dtype.capitalize(), name))


def _assert_file_not_exists(file_dir, dtype, name):
    if _file_exists(file_dir, dtype, name):
        raise SmifDataExistsError("%s '%s' already exists" % (dtype.capitalize(), name))


def _file_exists(file_dir, dtype, name):
    dir_key = "%ss" % dtype
    try:
        return os.path.exists(os.path.join(file_dir[dir_key], name + '.yml'))
    except TypeError:
        msg = "Could not parse file name {} and dtype {}"
        raise SmifDataNotFoundError(msg.format(name, dtype))


def _read_filenames_in_dir(path, extension):
    """Returns the name of the Yaml files in a certain directory

    Arguments
    ---------
    path: str
        Path to directory
    extension: str
        Extension of files (such as: '.yml' or '.csv')

    Returns
    -------
    list
        The list of files in `path` with extension
    """
    files = []
    for filename in os.listdir(path):
        if filename.endswith(extension):
            files.append(os.path.splitext(filename)[0])
    return files


def _skip_coords(config, keys):
    """Given a config dict and list of top-level keys for lists of specs,
    delete coords from each spec in each list.
    """
    config = copy.deepcopy(config)
    for key in keys:
        for spec in config[key]:
            try:
                del spec['coords']
            except KeyError:
                pass
    return config


def _pick_from_list(list_of_dicts, name):
    for item in list_of_dicts:
        if 'name' in item and item['name'] == name:
            return item
    return None


def _idx_in_list(list_of_dicts, name):
    for i, item in enumerate(list_of_dicts):
        if 'name' in item and item['name'] == name:
            return i
    return None



"""The store provides a common data interface to smif configuration, data and metadata.

Raises
------
SmifDataNotFoundError
    If data cannot be found in the store when try to read from the store
SmifDataExistsError
    If data already exists in the store when trying to write to the store
    (use an update method instead)
SmifDataMismatchError
    Data presented to read, write and update methods is in the
    incorrect format or of wrong dimensions to that expected
SmifDataReadError
    When unable to read data e.g. unable to handle file type or connect
    to database
"""
from copy import deepcopy
from logging import getLogger
from operator import itemgetter

import numpy as np  # type: ignore
from smif.data_layer import DataArray
from smif.data_layer.file import CSVDataStore, ParquetDataStore
from smif.data_layer.validate import (validate_sos_model_config,
                                      validate_sos_model_format)
from smif.exception import SmifDataNotFoundError
from smif.metadata.spec import Spec


class Store():
    """Common interface to data store, composed of config, metadata and data store implementations.

    Parameters
    ----------
    config_store: ~smif.data_layer.abstract_config_store.ConfigStore
    metadata_store: ~smif.data_layer.abstract_metadata_store.MetadataStore
    data_store: ~smif.data_layer.abstract_data_store.DataStore
    """
    def __init__(self, config_store, metadata_store, data_store, model_base_folder="."):
        self.logger = getLogger(__name__)
        self.config_store = config_store
        self.metadata_store = metadata_store
        self.data_store = data_store
        # base folder for any relative paths to models
        self.model_base_folder = str(model_base_folder)

    #
    # CONFIG
    #

    # region Model runs
    def read_model_runs(self):
        """Read all system-of-system model runs

        Returns
        -------
        list[~smif.controller.modelrun.ModelRun]
        """
        return sorted(self.config_store.read_model_runs(), key=itemgetter('name'))

    def read_model_run(self, model_run_name):
        """Read a system-of-system model run

        Parameters
        ----------
        model_run_name : str

        Returns
        -------
        ~smif.controller.modelrun.ModelRun
        """
        return self.config_store.read_model_run(model_run_name)

    def write_model_run(self, model_run):
        """Write system-of-system model run

        Parameters
        ----------
        model_run : ~smif.controller.modelrun.ModelRun
        """
        self.config_store.write_model_run(model_run)

    def update_model_run(self, model_run_name, model_run):
        """Update system-of-system model run

        Parameters
        ----------
        model_run_name : str
        model_run : ~smif.controller.modelrun.ModelRun
        """
        self.config_store.update_model_run(model_run_name, model_run)

    def delete_model_run(self, model_run_name):
        """Delete a system-of-system model run

        Parameters
        ----------
        model_run_name : str
        """
        self.config_store.delete_model_run(model_run_name)
    # endregion

    # region System-of-systems models
    def read_sos_models(self):
        """Read all system-of-system models

        Returns
        -------
        list[~smif.model.sos_model.SosModel]
        """
        return sorted(self.config_store.read_sos_models(), key=itemgetter('name'))

    def read_sos_model(self, sos_model_name):
        """Read a specific system-of-system model

        Parameters
        ----------
        sos_model_name : str

        Returns
        -------
        ~smif.model.sos_model.SosModel
        """
        return self.config_store.read_sos_model(sos_model_name)

    def write_sos_model(self, sos_model):
        """Write system-of-system model

        Parameters
        ----------
        sos_model : ~smif.model.sos_model.SosModel
        """
        validate_sos_model_format(sos_model)
        self.config_store.write_sos_model(sos_model)

    def update_sos_model(self, sos_model_name, sos_model):
        """Update system-of-system model

        Parameters
        ----------
        sos_model_name : str
        sos_model : ~smif.model.sos_model.SosModel
        """
        models = self.config_store.read_models()
        scenarios = self.config_store.read_scenarios()
        validate_sos_model_config(sos_model, models, scenarios)

        self.config_store.update_sos_model(sos_model_name, sos_model)

    def delete_sos_model(self, sos_model_name):
        """Delete a system-of-system model

        Parameters
        ----------
        sos_model_name : str
        """
        self.config_store.delete_sos_model(sos_model_name)
    # endregion

    # region Models
    def read_models(self, skip_coords=False):
        """Read all models

        Returns
        -------
        list[~smif.model.model.Model]
        """
        models = sorted(self.config_store.read_models(), key=itemgetter('name'))
        if not skip_coords:
            models = [
                self._add_coords(model, ('inputs', 'outputs', 'parameters'))
                for model in models
            ]
        return models

    def read_model(self, model_name, skip_coords=False):
        """Read a model

        Parameters
        ----------
        model_name : str

        Returns
        -------
        ~smif.model.model.Model
        """
        model = self.config_store.read_model(model_name)
        if not skip_coords:
            model = self._add_coords(model, ('inputs', 'outputs', 'parameters'))
        return model

    def write_model(self, model):
        """Write a model

        Parameters
        ----------
        model : ~smif.model.model.Model
        """
        self.config_store.write_model(model)

    def update_model(self, model_name, model):
        """Update a model

        Parameters
        ----------
        model_name : str
        model : ~smif.model.model.Model
        """
        self.config_store.update_model(model_name, model)

    def delete_model(self, model_name):
        """Delete a model

        Parameters
        ----------
        model_name : str
        """
        self.config_store.delete_model(model_name)
    # endregion

    # region Scenarios
    def read_scenarios(self, skip_coords=False):
        """Read scenarios

        Returns
        -------
        list[~smif.model.ScenarioModel]
        """
        scenarios = sorted(self.config_store.read_scenarios(), key=itemgetter('name'))
        if not skip_coords:
            scenarios = [
                self._add_coords(scenario, ['provides'])
                for scenario in scenarios
            ]
        return scenarios

    def read_scenario(self, scenario_name, skip_coords=False):
        """Read a scenario

        Parameters
        ----------
        scenario_name : str

        Returns
        -------
        ~smif.model.ScenarioModel
        """
        scenario = self.config_store.read_scenario(scenario_name)
        if not skip_coords:
            scenario = self._add_coords(scenario, ['provides'])
        return scenario

    def write_scenario(self, scenario):
        """Write scenario

        Parameters
        ----------
        scenario : ~smif.model.ScenarioModel
        """
        self.config_store.write_scenario(scenario)

    def update_scenario(self, scenario_name, scenario):
        """Update scenario

        Parameters
        ----------
        scenario_name : str
        scenario : ~smif.model.ScenarioModel
        """
        self.config_store.update_scenario(scenario_name, scenario)

    def delete_scenario(self, scenario_name):
        """Delete scenario from project configuration

        Parameters
        ----------
        scenario_name : str
        """
        self.config_store.delete_scenario(scenario_name)
    # endregion

    # region Scenario Variants
    def read_scenario_variants(self, scenario_name):
        """Read variants of a given scenario

        Parameters
        ----------
        scenario_name : str

        Returns
        -------
        list[dict]
        """
        scenario_variants = self.config_store.read_scenario_variants(scenario_name)
        return sorted(scenario_variants, key=itemgetter('name'))

    def read_scenario_variant(self, scenario_name, variant_name):
        """Read a scenario variant

        Parameters
        ----------
        scenario_name : str
        variant_name : str

        Returns
        -------
        dict
        """
        return self.config_store.read_scenario_variant(scenario_name, variant_name)

    def write_scenario_variant(self, scenario_name, variant):
        """Write scenario to project configuration

        Parameters
        ----------
        scenario_name : str
        variant : dict
        """
        self.config_store.write_scenario_variant(scenario_name, variant)

    def update_scenario_variant(self, scenario_name, variant_name, variant):
        """Update scenario to project configuration

        Parameters
        ----------
        scenario_name : str
        variant_name : str
        variant : dict
        """
        self.config_store.update_scenario_variant(scenario_name, variant_name, variant)

    def delete_scenario_variant(self, scenario_name, variant_name):
        """Delete scenario from project configuration

        Parameters
        ----------
        scenario_name : str
        variant_name : str
        """
        self.config_store.delete_scenario_variant(scenario_name, variant_name)
    # endregion

    # region Narratives
    def read_narrative(self, sos_model_name, narrative_name):
        """Read narrative from sos_model

        Parameters
        ----------
        sos_model_name : str
        narrative_name : str
        """
        return self.config_store.read_narrative(sos_model_name, narrative_name)
    # endregion

    # region Strategies
    def read_strategies(self, model_run_name):
        """Read strategies for a given model run

        Parameters
        ----------
        model_run_name : str

        Returns
        -------
        list[dict]
        """
        strategies = deepcopy(self.config_store.read_strategies(model_run_name))
        for strategy in strategies:
            if strategy['type'] == 'pre-specified-planning':
                strategy['interventions'] = self.data_store.read_strategy_interventions(
                    strategy)
        return strategies

    def write_strategies(self, model_run_name, strategies):
        """Write strategies for a given model_run

        Parameters
        ----------
        model_run_name : str
        strategies : list[dict]
        """
        self.config_store.write_strategies(model_run_name, strategies)
    # endregion

    #
    # METADATA
    #

    # region Units
    def read_unit_definitions(self):
        """Reads custom unit definitions

        Returns
        -------
        list[str]
            Pint-compatible unit definitions
        """
        return self.metadata_store.read_unit_definitions()

    def write_unit_definitions(self, definitions):
        """Reads custom unit definitions

        Parameters
        ----------
        list[str]
            Pint-compatible unit definitions
        """
        self.metadata_store.write_unit_definitions(definitions)
    # endregion

    # region Dimensions
    def read_dimensions(self, skip_coords=False):
        """Read dimensions

        Returns
        -------
        list[~smif.metadata.coords.Coords]
        """
        return self.metadata_store.read_dimensions(skip_coords)

    def read_dimension(self, dimension_name, skip_coords=False):
        """Return dimension

        Parameters
        ----------
        dimension_name : str

        Returns
        -------
        ~smif.metadata.coords.Coords
            A dimension definition (including elements)
        """
        return self.metadata_store.read_dimension(dimension_name, skip_coords)

    def write_dimension(self, dimension):
        """Write dimension to project configuration

        Parameters
        ----------
        dimension : ~smif.metadata.coords.Coords
        """
        self.metadata_store.write_dimension(dimension)

    def update_dimension(self, dimension_name, dimension):
        """Update dimension

        Parameters
        ----------
        dimension_name : str
        dimension : ~smif.metadata.coords.Coords
        """
        self.metadata_store.update_dimension(dimension_name, dimension)

    def delete_dimension(self, dimension_name):
        """Delete dimension

        Parameters
        ----------
        dimension_name : str
        """
        self.metadata_store.delete_dimension(dimension_name)

    def _add_coords(self, item, keys):
        """Add coordinates to spec definitions on an object
        """
        item = deepcopy(item)
        for key in keys:
            spec_list = item[key]
            for spec in spec_list:
                if 'dims' in spec and spec['dims']:
                    spec['coords'] = {
                        dim: self.read_dimension(dim)['elements']
                        for dim in spec['dims']
                    }
        return item
    # endregion

    #
    # DATA
    #

    # region Scenario Variant Data
    def read_scenario_variant_data(
            self, scenario_name, variant_name, variable, timestep=None) -> DataArray:
        """Read scenario data file

        Parameters
        ----------
        scenario_name : str
        variant_name : str
        variable : str
        timestep : int (optional)
            If None, read data for all timesteps

        Returns
        -------
        data : ~smif.data_layer.data_array.DataArray
        """
        variant = self.read_scenario_variant(scenario_name, variant_name)
        key = self._key_from_data(variant['data'][variable], scenario_name, variant_name,
                                  variable)
        scenario = self.read_scenario(scenario_name)
        spec_dict = _pick_from_list(scenario['provides'], variable)
        spec = Spec.from_dict(spec_dict)
        return self.data_store.read_scenario_variant_data(key, spec, timestep)

    def write_scenario_variant_data(self, scenario_name, variant_name, data, timestep=None):
        """Write scenario data file

        Parameters
        ----------
        scenario_name : str
        variant_name : str
        data : ~smif.data_layer.data_array.DataArray
        timestep : int (optional)
            If None, write data for all timesteps
        """
        variant = self.read_scenario_variant(scenario_name, variant_name)
        key = self._key_from_data(variant['data'][data.spec.name], scenario_name, variant_name,
                                  data.spec.name)
        self.data_store.write_scenario_variant_data(key, data, timestep)
    # endregion

    # region Narrative Data
    def read_narrative_variant_data(self, sos_model_name, narrative_name, variant_name,
                                    parameter_name, timestep=None):
        """Read narrative data file

        Parameters
        ----------
        sos_model_name : str
        narrative_name : str
        variant_name : str
        parameter_name : str
        timestep : int (optional)
            If None, read data for all timesteps

        Returns
        -------
        ~smif.data_layer.data_array.DataArray
        """
        sos_model = self.read_sos_model(sos_model_name)
        narrative = _pick_from_list(sos_model['narratives'], narrative_name)
        variant = _pick_from_list(narrative['variants'], variant_name)
        key = self._key_from_data(variant['data'][parameter_name], narrative_name,
                                  variant_name, parameter_name)

        spec_dict = None
        # find sector model which needs this parameter, to get spec definition
        for model_name, params in narrative['provides'].items():
            if parameter_name in params:
                sector_model = self.read_model(model_name)
                spec_dict = _pick_from_list(sector_model['parameters'], parameter_name)
                break
        # find spec
        if spec_dict is None:
            raise SmifDataNotFoundError("Parameter {} not found in any of {}".format(
                parameter_name, sos_model['sector_models']))
        spec = Spec.from_dict(spec_dict)

        return self.data_store.read_narrative_variant_data(key, spec, timestep)

    def write_narrative_variant_data(self, sos_model_name, narrative_name, variant_name,
                                     data, timestep=None):
        """Read narrative data file

        Parameters
        ----------
        sos_model_name : str
        narrative_name : str
        variant_name : str
        data : ~smif.data_layer.data_array.DataArray
        timestep : int (optional)
            If None, write data for all timesteps
        """
        sos_model = self.read_sos_model(sos_model_name)
        narrative = _pick_from_list(sos_model['narratives'], narrative_name)
        variant = _pick_from_list(narrative['variants'], variant_name)
        key = self._key_from_data(
            variant['data'][data.spec.name], narrative_name, variant_name, data.spec.name)
        self.data_store.write_narrative_variant_data(key, data)

    def read_model_parameter_default(self, model_name, parameter_name):
        """Read default data for a sector model parameter

        Parameters
        ----------
        model_name : str
        parameter_name : str

        Returns
        -------
        ~smif.data_layer.data_array.DataArray
        """
        model = self.read_model(model_name)
        param = _pick_from_list(model['parameters'], parameter_name)
        spec = Spec.from_dict(param)
        try:
            path = param['default']
        except TypeError:
            raise SmifDataNotFoundError("Parameter {} not found in model {}".format(
                parameter_name, model_name))
        except KeyError:
            path = 'default__{}__{}.csv'.format(model_name, parameter_name)
        key = self._key_from_data(path, model_name, parameter_name)
        return self.data_store.read_model_parameter_default(key, spec)

    def write_model_parameter_default(self, model_name, parameter_name, data):
        """Write default data for a sector model parameter

        Parameters
        ----------
        model_name : str
        parameter_name : str
        data : ~smif.data_layer.data_array.DataArray
        """
        model = self.read_model(model_name, skip_coords=True)
        param = _pick_from_list(model['parameters'], parameter_name)
        try:
            path = param['default']
        except TypeError:
            raise SmifDataNotFoundError("Parameter {} not found in model {}".format(
                parameter_name, model_name))
        except KeyError:
            path = 'default__{}__{}.csv'.format(model_name, parameter_name)
        key = self._key_from_data(path, model_name, parameter_name)
        self.data_store.write_model_parameter_default(key, data)
    # endregion

    # region Interventions
    def read_interventions(self, model_name):
        """Read interventions data for `model_name`

        Returns
        -------
        dict[str, dict]
            A dict of intervention dictionaries containing intervention
            attributes keyed by intervention name
        """
        model = self.read_model(model_name, skip_coords=True)
        if model['interventions'] != []:
            return self.data_store.read_interventions(model['interventions'])
        else:
            return {}

    def write_interventions(self, model_name, interventions):
        """Write interventions data for a model

        Parameters
        ----------
        dict[str, dict]
            A dict of intervention dictionaries containing intervention
            attributes keyed by intervention name
        """
        model = self.read_model(model_name)
        model['interventions'] = [model_name + '.csv']
        self.update_model(model_name, model)
        self.data_store.write_interventions(model['interventions'][0], interventions)

    def read_strategy_interventions(self, strategy):
        """Read interventions as defined in a model run strategy
        """
        return self.data_store.read_strategy_interventions(strategy)

    def read_initial_conditions(self, model_name):
        """Read historical interventions for `model_name`

        Returns
        -------
        list[dict]
            A list of historical interventions, with keys 'name' and 'build_year'
        """
        model = self.read_model(model_name)
        if model['initial_conditions'] != []:
            return self.data_store.read_initial_conditions(model['initial_conditions'])
        else:
            return []

    def write_initial_conditions(self, model_name, initial_conditions):
        """Write historical interventions for a model

        Parameters
        ----------
        list[dict]
            A list of historical interventions, with keys 'name' and 'build_year'
        """
        model = self.read_model(model_name)
        model['initial_conditions'] = [model_name + '.csv']
        self.update_model(model_name, model)
        self.data_store.write_initial_conditions(model['initial_conditions'][0],
                                                 initial_conditions)

    def read_all_initial_conditions(self, model_run_name):
        """A list of all historical interventions

        Returns
        -------
        list[dict]
        """
        historical_interventions = []
        model_run = self.read_model_run(model_run_name)
        sos_model_name = model_run['sos_model']
        sos_model = self.read_sos_model(sos_model_name)
        sector_model_names = sos_model['sector_models']
        for sector_model_name in sector_model_names:
            historical_interventions.extend(
                self.read_initial_conditions(sector_model_name)
            )
        return historical_interventions
    # endregion

    # region State
    def read_state(self, model_run_name, timestep, decision_iteration=None):
        """Read list of (name, build_year) for a given model_run, timestep,
        decision

        Parameters
        ----------
        model_run_name : str
        timestep : int
        decision_iteration : int, optional

        Returns
        -------
        list[dict]
        """
        return self.data_store.read_state(model_run_name, timestep, decision_iteration)

    def write_state(self, state, model_run_name, timestep, decision_iteration=None):
        """State is a list of decisions with name and build_year.

        State is output from the DecisionManager

        Parameters
        ----------
        state : list[dict]
        model_run_name : str
        timestep : int
        decision_iteration : int, optional
        """
        self.data_store.write_state(state, model_run_name, timestep, decision_iteration)
    # endregion

    # region Conversion coefficients
    def read_coefficients(self, source_dim: str, destination_dim: str) -> np.ndarray:
        """Reads coefficients from the store

        Coefficients are uniquely identified by their source/destination dimensions.
        This method and `write_coefficients` implement caching of conversion
        coefficients between dimensions.

        Parameters
        ----------
        source_dim : str
            Dimension name
        destination_dim : str
            Dimension name

        Returns
        -------
        numpy.ndarray

        Notes
        -----
        To be called from :class:`~smif.convert.adaptor.Adaptor` implementations.
        """
        return self.data_store.read_coefficients(source_dim, destination_dim)

    def write_coefficients(self, source_dim: str, destination_dim: str, data: np.ndarray):
        """Writes coefficients to the store

        Coefficients are uniquely identified by their source/destination dimensions.
        This method and `read_coefficients` implement caching of conversion
        coefficients between dimensions.

        Parameters
        ----------
        source_dim : str
            Dimension name
        destination_dim : str
            Dimension name
        data : numpy.ndarray

        Notes
        -----
        To be called from :class:`~smif.convert.adaptor.Adaptor` implementations.
        """
        self.data_store.write_coefficients(source_dim, destination_dim, data)
    # endregion

    # region Results
    def read_results(self, model_run_name, model_name, output_spec, timestep=None,
                     decision_iteration=None) -> DataArray:
        """Return results of a `model_name` in `model_run_name` for a given `output_name`

        Parameters
        ----------
        model_run_id : str
        model_name : str
        output_spec : smif.metadata.Spec
        timestep : int, default=None
        decision_iteration : int, default=None

        Returns
        -------
        ~smif.data_layer.data_array.DataArray
        """
        return self.data_store.read_results(
            model_run_name, model_name, output_spec, timestep, decision_iteration)

    def write_results(self, data_array, model_run_name, model_name, timestep=None,
                      decision_iteration=None):
        """Write results of a `model_name` in `model_run_name` for a given `output_name`

        Parameters
        ----------
        data_array : ~smif.data_layer.data_array.DataArray
        model_run_id : str
        model_name : str
        timestep : int, optional
        decision_iteration : int, optional
        """
        self.data_store.write_results(
            data_array, model_run_name, model_name, timestep, decision_iteration)

    def available_results(self, model_run_name):
        """List available results from a model run

        Returns
        -------
        list[tuple]
             Each tuple is (timestep, decision_iteration, model_name, output_name)
        """
        return self.data_store.available_results(model_run_name)

    def prepare_warm_start(self, model_run_name):
        """Copy the results from the previous model_run if available

        The method allows a previous unsuccessful model_run to 'warm start' a new model run
        from a later timestep. Model results are recovered from the timestep that the previous
        model_run was run until, and the new model run runs from the returned timestep

        Parameters
        ----------
        model_run_name : str

        Returns
        -------
        int The timestep to which the data store was recovered

        Notes
        -----
        Called from smif.controller.execute
        """
        available_results = self.available_results(model_run_name)
        if available_results:
            max_timestep = max(
                timestep for
                timestep, decision_iteration, model_name, output_name in available_results
            )
            # could explicitly clear results for max timestep
        else:
            max_timestep = None
        return max_timestep
    # endregion

    # region data store utilities
    def _key_from_data(self, path, *args):
        """Return path or generate a unique key for a given set of args
        """
        if isinstance(self.data_store, (CSVDataStore, ParquetDataStore)):
            return path
        else:
            return tuple(args)
    # endregion


def _pick_from_list(list_of_dicts, name):
    for item in list_of_dicts:
        if 'name' in item and item['name'] == name:
            return item
    return None
